import logging

import click
import numpy as np

import vpype as vp
import vpype_viewer
from vpype_viewer import ViewMode

from .cli import cli

__all__ = ["show"]

COLORS = [
    (0, 0, 1),
    (0, 0.5, 0),
    (1, 0, 0),
    (0, 0.75, 0.75),
    (0, 1, 0),
    (0.75, 0, 0.75),
    (0.75, 0.75, 0),
    (0, 0, 0),
]


@cli.command(group="Output")
@click.option("--classic", is_flag=True, help="Use the classic viewer.")
@click.option("--force", is_flag=True, hidden=True, help="Force modern viewer")
@click.option("-p", "--show-pen-up", is_flag=True, help="Display pen-up trajectories.")
@click.option("-d", "--show-points", is_flag=True, help="Display paths points with dots.")
@click.option("-o", "--outline", is_flag=True, help="Display in outline mode (modern only).")
@click.option(
    "-c",
    "--colorful",
    is_flag=True,
    help="Display in outline colorful mode (takes precedence over --outline).",
)
@click.option("-a", "--show-axes", is_flag=True, help="Display axes (classic only).")
@click.option(
    "-g", "--show-grid", is_flag=True, help="Display grid (implies -a, classic only)."
)
@click.option(
    "-h", "--hide-legend", is_flag=True, help="Do not display the legend (classic only)."
)
@click.option(
    "-u",
    "--unit",
    type=str,
    default="px",
    help="Units of the plot (when --show-grid is active, classic only).",
)
@vp.global_processor
def show(
    document: vp.Document,
    classic: bool,
    force: bool,
    show_pen_up: bool,
    show_points: bool,
    outline: bool,
    colorful: bool,
    show_axes: bool,
    show_grid: bool,
    hide_legend: bool,
    unit: str,
):
    """Display the geometry in an graphical user interface.

    By default, this command use a modern, hardware-accelerated viewer (currently in beta) with
    a preview mode (adjustable pen width and opacity) and interactive controls to adjust
    display options. This viewer requires OpenGL 3.3 support.

    The original, Matplotlib-based viewer is still available with the `--classic` option. The
    classic viewer does not have interactive controls for display options. Use the command-line
    options to customize the display.
    """

    if not classic:
        mgl_ok = _test_mgl()
        if not mgl_ok and not force:
            classic = True
            logging.warning("!!! show: ModernGL not available, reverting to classic mode.")
        elif not mgl_ok and force:
            logging.warning("!!! show: ModernGL not available but forced to modern mode.")

    if classic:
        _show_mpl(
            document,
            show_axes,
            show_grid,
            show_pen_up,
            show_points,
            hide_legend,
            colorful,
            unit,
        )
    else:
        view_mode = ViewMode.PREVIEW
        if outline or show_points:
            view_mode = ViewMode.OUTLINE
        if colorful:
            view_mode = ViewMode.OUTLINE_COLORFUL

        vpype_viewer.show(
            document, view_mode=view_mode, show_pen_up=show_pen_up, show_points=show_points
        )

    return document


def _test_mgl() -> bool:
    """Tests availability of ModernGL."""
    # noinspection PyBroadException
    try:
        import glcontext

        backend = glcontext.default_backend()
        ctx = backend(mode="standalone", glversion=330)
        if ctx.load("glProgramUniform1iv") == 0:
            logging.info("ModernGL detection: glProgramUniform1iv not found")
            return False
    except Exception as exc:
        logging.info(f"ModernGL detection failed with error {exc}")
        return False

    return True


def _show_mpl(
    document: vp.Document,
    show_axes: bool,
    show_grid: bool,
    show_pen_up: bool,
    show_points: bool,
    hide_legend: bool,
    colorful: bool,
    unit: str,
):
    """Display the geometry using matplotlib.

    By default, only the geometries are displayed without the axis. All geometries are
    displayed with black. When using the `--colorful` flag, each segment will have a different
    color (default matplotlib behaviour). This can be useful for debugging purposes.
    """

    # deferred import to optimise startup time
    import matplotlib.collections
    import matplotlib.pyplot as plt

    scale = 1 / vp.convert_length(unit)

    fig = plt.figure()
    color_idx = 0
    collections = {}

    # draw page boundaries
    if document.page_size:
        w = document.page_size[0] * scale
        h = document.page_size[1] * scale
        dw = 10 * scale
        plt.plot(
            np.array([0, 1, 1, 0, 0]) * w,
            np.array([0, 0, 1, 1, 0]) * h,
            "-k",
            lw=0.25,
            label=None,
        )
        plt.fill(
            np.array([w, w + dw, w + dw, dw, dw, w]),
            np.array([dw, dw, h + dw, h + dw, h, h]),
            "k",
            alpha=0.3,
            label=None,
        )

    for layer_id, lc in document.layers.items():
        if colorful:
            color = COLORS[color_idx:] + COLORS[:color_idx]
            marker_color = "k"
            color_idx += len(lc)
        else:
            color = COLORS[color_idx]  # type: ignore
            marker_color = [color]  # type: ignore
            color_idx += 1
        if color_idx >= len(COLORS):
            color_idx = color_idx % len(COLORS)

        layer_lines = matplotlib.collections.LineCollection(
            (vp.as_vector(line) * scale for line in lc),
            color=color,
            lw=1,
            alpha=0.5,
            label=str(layer_id),
        )
        collections[layer_id] = [layer_lines]
        plt.gca().add_collection(layer_lines)

        if show_points:
            points = np.hstack([line for line in lc]) * scale
            layer_points = plt.gca().scatter(
                points.real, points.imag, marker=".", c=marker_color, s=16
            )
            collections[layer_id].append(layer_points)

        if show_pen_up:
            pen_up_lines = matplotlib.collections.LineCollection(
                (
                    (vp.as_vector(lc[i])[-1] * scale, vp.as_vector(lc[i + 1])[0] * scale)
                    for i in range(len(lc) - 1)
                ),
                color=(0, 0, 0),
                lw=0.5,
                alpha=0.5,
            )
            collections[layer_id].append(pen_up_lines)
            plt.gca().add_collection(pen_up_lines)

    plt.gca().invert_yaxis()
    plt.axis("equal")
    plt.margins(0, 0)

    if not hide_legend:
        lgd = plt.legend(loc="upper right")
        # we will set up a dict mapping legend line to orig line, and enable
        # picking on the legend line
        line_dict = {}
        for lgd_line, lgd_text in zip(lgd.get_lines(), lgd.get_texts()):
            lgd_line.set_picker(True)  # 5 pts tolerance
            lgd_line.set_pickradius(5)
            layer_id = int(lgd_text.get_text())
            if layer_id in collections:
                line_dict[lgd_line] = collections[layer_id]

        def on_pick(event):
            line = event.artist
            vis = not line_dict[line][0].get_visible()
            for ln in line_dict[line]:
                ln.set_visible(vis)

            if vis:
                line.set_alpha(1.0)
            else:
                line.set_alpha(0.2)
            fig.canvas.draw()

        fig.canvas.mpl_connect("pick_event", on_pick)

    if show_axes or show_grid:
        plt.axis("on")
        plt.xlabel(f"[{unit}]")
        plt.ylabel(f"[{unit}]")
    else:
        plt.axis("off")
    if show_grid:
        plt.grid("on")
    plt.show()



import logging
import math
from typing import List, Optional, Tuple, Union, cast

import click

import vpype as vp

from .cli import cli

__all__ = ("rotate", "scale_relative", "scaleto", "skew", "translate")


def _compute_origin(
    document: vp.Document,
    layer: Optional[Union[int, List[int]]],
    origin_coords: Union[Tuple[()], Tuple[float, float]],
) -> Tuple[Tuple[float, float], List[int], Tuple[float, float, float, float]]:
    layer_ids = vp.multiple_to_layer_ids(layer, document)
    bounds = document.bounds(layer_ids)

    if not bounds:
        logging.warning("no geometry available, cannot compute origin")
        raise ValueError

    if len(origin_coords) == 2:
        origin = origin_coords
    else:
        origin = (
            0.5 * (bounds[0] + bounds[2]),
            0.5 * (bounds[1] + bounds[3]),
        )

    return cast(Tuple[float, float], origin), layer_ids, bounds


@cli.command(group="Transforms")
@click.argument("offset", nargs=2, type=vp.LengthType(), required=True)
@vp.layer_processor
def translate(lc: vp.LineCollection, offset: Tuple[float, float]):
    """
    Translate the geometries. X and Y offsets must be provided. These arguments understand
    supported units.
    """
    lc.translate(offset[0], offset[1])
    return lc


# noinspection PyShadowingNames
@cli.command(name="scale", group="Transforms")
@click.argument("scale", nargs=2, type=vp.LengthType())
@click.option(
    "-l",
    "--layer",
    type=vp.LayerType(accept_multiple=True),
    default="all",
    help="Target layer(s).",
)
@click.option(
    "-o",
    "--origin",
    "origin_coords",
    nargs=2,
    type=vp.LengthType(),
    help="Use a specific origin.",
)
@vp.global_processor
def scale_relative(
    document: vp.Document,
    scale: Tuple[float, float],
    layer: Union[int, List[int]],
    origin_coords: Union[Tuple[()], Tuple[float, float]],
):
    """Scale the geometries by a factor.

    The origin used is the bounding box center, unless the `--origin` option is used.

    By default, act on all layers. If one or more layer IDs are provided with the `--layer`
    option, only these layers will be affected. In this case, the bounding box is that of the
    listed layers.

    Example:

        Double the size of the geometries in layer 1, using (0, 0) as origin:

            vpype [...] scale -l 1 -o 0 0 2 2 [...]

    """

    try:
        origin, layer_ids, _ = _compute_origin(document, layer, origin_coords)
    except ValueError:
        return document

    for vid in layer_ids:
        lc = document[vid]
        lc.translate(-origin[0], -origin[1])
        lc.scale(scale[0], scale[1])
        lc.translate(origin[0], origin[1])

    return document


# noinspection PyShadowingNames
@cli.command(group="Transforms")
@click.argument("dim", nargs=2, type=vp.LengthType())
@click.option(
    "-l",
    "--layer",
    type=vp.LayerType(accept_multiple=True),
    default="all",
    help="Target layer(s).",
)
@click.option(
    "-f",
    "--fit-dimensions",
    is_flag=True,
    help="Exactly fit target dimension, distorting geometries if required.",
)
@click.option(
    "-o",
    "--origin",
    "origin_coords",
    nargs=2,
    type=vp.LengthType(),
    help="Use a specific origin.",
)
@vp.global_processor
def scaleto(
    document: vp.Document,
    dim: Tuple[float, float],
    layer: Union[int, List[int]],
    fit_dimensions: bool,
    origin_coords: Union[Tuple[()], Tuple[float, float]],
):
    """Scale the geometries to given dimensions.

    By default, the homogeneous scaling is applied on both X and Y directions, even if the
    geometry proportions are not the same as the target dimensions.  When passing
    `--fit-dimensions`, the geometries are scaled such as to fit exactly the target dimensions,
    distorting them if required.

    The origin used is the bounding box center, unless the `--origin` option is used.

    By default, act on all layers. If one or more layer IDs are provided with the `--layer`
    option, only these layers will be affected. In this case, the bounding box is that of the
    listed layers.

    Example:

        Scale a SVG to a A4 page, accounting for 1cm margin:

            vpype read input.svg scaleto 19cm 27.7cm write -p a4 -c output.svg

    """

    try:
        origin, layer_ids, bounds = _compute_origin(document, layer, origin_coords)
    except ValueError:
        return document

    width = bounds[2] - bounds[0]
    height = bounds[3] - bounds[1]

    if width == 0.0 or height == 0.0:
        return document

    factors = dim[0] / width, dim[1] / height
    if not fit_dimensions:
        factors = (min(factors), min(factors))

    for vid in layer_ids:
        lc = document[vid]
        lc.translate(-origin[0], -origin[1])
        lc.scale(factors[0], factors[1])
        lc.translate(origin[0], origin[1])

    return document


# noinspection DuplicatedCode
@cli.command(group="Transforms")
@click.argument("angle", required=True, type=vp.AngleType())
@click.option(
    "-l",
    "--layer",
    type=vp.LayerType(accept_multiple=True),
    default="all",
    help="Target layer(s).",
)
@click.option(
    "-o",
    "--origin",
    "origin_coords",
    nargs=2,
    type=vp.LengthType(),
    help="Use a specific origin.",
)
@vp.global_processor
def rotate(
    document: vp.Document,
    angle: float,
    layer: Union[int, List[int]],
    origin_coords: Union[Tuple[()], Tuple[float, float]],
):
    """Rotate the geometries (clockwise positive).

    The origin used is the bounding box center, unless the `--origin` option is used.

    ANGLE is in degrees by default, but alternative CSS unit may be provided.

    By default, act on all layers. If one or more layer IDs are provided with the `--layer`
    option, only these layers will be affected. In this case, the bounding box is that of the
    listed layers.
    """

    try:
        origin, layer_ids, _ = _compute_origin(document, layer, origin_coords)
    except ValueError:
        return document

    for vid in layer_ids:
        lc = document[vid]
        lc.translate(-origin[0], -origin[1])
        lc.rotate(angle * math.pi / 180.0)
        lc.translate(origin[0], origin[1])

    return document


# noinspection DuplicatedCode
@cli.command(group="Transforms")
@click.argument("angles", required=True, nargs=2, type=float)
@click.option(
    "-l",
    "--layer",
    type=vp.LayerType(accept_multiple=True),
    default="all",
    help="Target layer(s).",
)
@click.option(
    "-o",
    "--origin",
    "origin_coords",
    nargs=2,
    type=vp.LengthType(),
    help="Use a specific origin.",
)
@vp.global_processor
def skew(
    document: vp.Document,
    layer: Union[int, List[int]],
    angles: Tuple[float, float],
    origin_coords: Union[Tuple[()], Tuple[float, float]],
):
    """Skew the geometries.

    The geometries are sheared by the provided angles along X and Y dimensions.

    ANGLE is in degrees by default, but alternative CSS unit may be provided.

    The origin used in the bounding box center, unless the `--centroid` or `--origin` options
    are used.
    """

    try:
        origin, layer_ids, _ = _compute_origin(document, layer, origin_coords)
    except ValueError:
        return document

    for vid in layer_ids:
        lc = document[vid]
        lc.translate(-origin[0], -origin[1])
        lc.skew(angles[0] * math.pi / 180.0, angles[1] * math.pi / 180.0)
        lc.translate(origin[0], origin[1])

    return document



import logging
import os
import random
import shlex
from typing import Any, List, Optional, TextIO, Union

import click
import numpy as np
from click import get_os_args
from click_plugins import with_plugins
from pkg_resources import iter_entry_points
from shapely.geometry import MultiLineString

import vpype as vp

__all__ = ("cli", "execute", "begin", "end")


class GroupedGroup(click.Group):
    """Custom group class which implements command grouping in --help display.

    Based on Stephen Rauch's excellent answer: https://stackoverflow.com/a/58770064/229511
    """

    def command(self, *args, **kwargs):
        """Gather the command help groups"""
        help_group = kwargs.pop("group", None)
        decorator = super(GroupedGroup, self).command(*args, **kwargs)

        def wrapper(f):
            cmd = decorator(f)
            cmd.help_group = help_group
            return cmd

        return wrapper

    def format_commands(self, ctx, formatter):
        """Extra format methods for multi methods that adds all the commands
        after the options.
        """
        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            # What is this, the tool lied about a command.  Ignore it
            if cmd is None:
                continue
            if cmd.hidden:
                continue

            commands.append((subcommand, cmd))

        # allow for 3 times the default spacing
        if len(commands):
            longest = max(len(cmd[0]) for cmd in commands)
            limit = formatter.width - 6 - longest

            groups = {}
            for subcommand, cmd in commands:
                help_text = cmd.get_short_help_str(limit)
                subcommand += " " * (longest - len(subcommand))
                groups.setdefault(getattr(cmd, "help_group", "Unknown"), []).append(
                    (subcommand, help_text)
                )

            with formatter.section("Commands"):
                for group_name, rows in groups.items():
                    with formatter.section(group_name):
                        formatter.write_dl(rows)

    def main(self, args=None, **extra):
        """Let's get a chance to pre-process the argument list for include options."""
        if args is None:
            args = get_os_args()
        return super().main(args=preprocess_argument_list(args), **extra)


# noinspection PyUnusedLocal,PyUnresolvedReferences
@with_plugins(iter_entry_points("vpype.plugins"))
@click.group(cls=GroupedGroup, chain=True)
@click.version_option(version=vp.__version__, message="%(prog)s %(version)s")
@click.option("-v", "--verbose", count=True)
@click.option("-I", "--include", type=click.Path(), help="Load commands from a command file.")
@click.option(
    "-H",
    "--history",
    is_flag=True,
    help="Record this command in a `vpype_history.txt` in the current directory.",
)
@click.option("-s", "--seed", type=int, help="Specify the RNG seed.")
@click.option(
    "-c", "--config", type=click.Path(exists=True), help="Load an additional config file."
)
@click.pass_context
def cli(ctx, verbose, include, history, seed, config):
    """Execute the vector processing pipeline passed as argument."""

    logging.basicConfig()
    if verbose == 0:
        logging.getLogger().setLevel(logging.WARNING)
    elif verbose == 1:
        logging.getLogger().setLevel(logging.INFO)
    elif verbose > 1:
        logging.getLogger().setLevel(logging.DEBUG)

    # We use the command string as context object, mainly for the purpose of the `write`
    # command. This is a bit of a hack, and will need to be updated if we ever need more state
    # to be passed around (probably VpypeState should go in there!)
    cmd_string = "vpype " + " ".join(shlex.quote(arg) for arg in get_os_args()) + "\n"
    ctx.obj = cmd_string

    if history:
        with open("vpype_history.txt", "a") as fp:
            fp.write(cmd_string)

    if seed is None:
        seed = np.random.randint(2 ** 31)
        logging.info(f"vpype: no seed provided, using {seed}")
    np.random.seed(seed)
    random.seed(seed)

    if config is not None:
        vp.config_manager.load_config_file(config)


# noinspection PyShadowingNames,PyUnusedLocal
@cli.resultcallback()
def process_pipeline(processors, verbose, include, history, seed, config):
    execute_processors(processors)


def execute_processors(processors) -> vp.VpypeState:
    """
    Execute a sequence of processors to generate a Document structure. For block handling, we
    use a recursive approach. Only top-level blocks are extracted and processed by block
    processors, which, in turn, recursively call this function.
    :param processors: iterable of processors
    :return: generated geometries
    """

    outer_processors: List[Any] = []  # gather commands outside of top-level blocks
    top_level_processors: List[Any] = []  # gather commands inside of top-level blocks
    block = None  # save the current top-level block's block layer_processor
    nested_count = 0  # block depth counter
    expect_block = False  # set to True by `begin` command

    for proc in processors:
        if isinstance(proc, BlockProcessor):
            if expect_block:
                expect_block = False
                # if we in a top level block, we save the block layer_processor
                # (nested block are ignored for the time being)
                if nested_count == 1:
                    block = proc
                else:
                    top_level_processors.append(proc)
            else:
                raise click.ClickException("A block command must always follow 'begin'")
        elif expect_block:
            raise click.ClickException("A block command must always follow 'begin'")
        elif isinstance(proc, BeginBlock):
            # entering a block
            nested_count += 1
            expect_block = True

            if nested_count > 1:
                top_level_processors.append(proc)
        elif isinstance(proc, EndBlock):
            if nested_count < 1:
                raise click.ClickException(
                    "A 'end' command has no corresponding 'begin' command"
                )

            nested_count -= 1

            if nested_count == 0:
                # we're closing a top level block, let's process it
                block_document = block.process(top_level_processors)  # type: ignore

                # Create a placeholder layer_processor that will add the block's result to the
                # current frame. The placeholder_processor is a closure, so we need to make
                # a closure-building function. Failing that, the closure would refer directly
                # to the block_vd variable above, which might be overwritten by a subsequent
                # top-level block
                # noinspection PyShadowingNames
                def build_placeholder_processor(block_document):
                    def placeholder_processor(input_state):
                        input_state.document.extend(block_document)
                        return input_state

                    return placeholder_processor

                outer_processors.append(build_placeholder_processor(block_document))

                # reset the top level layer_processor list
                top_level_processors = list()
            else:
                top_level_processors.append(proc)
        else:
            # this is a 'normal' layer_processor, we can just add it to the top of the stack
            if nested_count == 0:
                outer_processors.append(proc)
            else:
                top_level_processors.append(proc)

    # at this stage, the stack must have a single frame, otherwise we're missing end commands
    if nested_count > 0:
        raise click.ClickException("An 'end' command is missing")

    # the (only) frame's processors should now be flat and can be chain-called
    state = vp.VpypeState()
    for proc in outer_processors:
        state = proc(state)
    return state


class BeginBlock:
    pass


@cli.command(group="Block control")
def begin():
    """Marks the start of a block.

    A `begin` command must be followed by a block processor command (eg. `grid` or `repeat`),
    which indicates how the block is processed. Blocks must be ended by a `end` command.

    Blocks can be nested.
    """
    return BeginBlock()


class EndBlock:
    pass


@cli.command(group="Block control")
def end():
    """Marks the end of a block."""
    return EndBlock()


class BlockProcessor:
    """
    Base class for all block processors. Although it does nothing, block processors must
    sub-class :class:`BlockProcessor` to be recognized as such.
    """

    def process(self, processors) -> MultiLineString:
        """
        Generate the compound geometries based on the provided processors. Sub-class must
        override this function in their implementation.
        :param processors: list of processors
        :return: compound geometries
        """


def extract_arguments(f: TextIO) -> List[str]:
    """Read the content of a file-like object and extract the corresponding argument list.

    Everything following a '#' is ignored until end of line. Any whitespace is considered
    to separate arguments. Single and double quote are honored, i.e. content becomes an
    argument (but quote are removed).

    :param f: file-like object
    :return: list of argument extracted from input
    """
    args = []
    for line in f.readlines():
        args.extend(shlex.split(line, comments=True))
    return args


def preprocess_argument_list(args: List[str], cwd: Union[str, None] = None) -> List[str]:
    """Preprocess an argument list, replacing 'include' options by the corresponding file's
    content.

    Include options are either '-I' or '--include', and must be followed by a file path. This
    behaviour is recursive, e.g. a file could contain an include statement as well.

    :param args:
    :param cwd: current working directory, used as reference for relative file paths (use
        actual current working directory if None)
    :return: preprocessed list or argument
    """

    if cwd is None:
        cwd = os.getcwd()

    result = []

    while len(args) > 0:
        arg = args.pop(0)

        if arg == "-I" or arg == "--include":
            if len(args) == 0:
                raise click.ClickException("include option must be followed by a file path")
            else:
                # include statement in files are relative to that file, so we need to
                # provide the file's path to ourselves
                file_path = args.pop(0)
                if not os.path.isabs(file_path):
                    file_path = os.path.join(cwd, file_path)
                dir_path = os.path.dirname(file_path)

                with open(file_path, "r") as f:
                    result.extend(preprocess_argument_list(extract_arguments(f), dir_path))
        else:
            result.append(arg)

    return result


def execute(pipeline: str, document: Optional[vp.Document] = None) -> vp.Document:
    """Execute a vpype pipeline.

    This function serves as a Python API to vpype's pipeline. It can be used from a regular
    Python script (as opposed to the ``vpype`` CLI which must be used from a console or via
    :func:`os.system`.

    If a :class:`vpype.Document` instance is provided, it will be preloaded in the pipeline
    before the first command executes. The pipeline's content after the last command is
    returned as a :class:`vpype.Document` instance.

    Examples:

        Read a SVG file, optimize it and return the result as a :class:`vpype.Document`
        instance::

            >>> doc = execute("read input.svg linemerge linesimplify linesort")

        Optimize and save a :class:`vpype.Document` instance::

            >>> doc = vp.Document()
            >>> # populate `doc` with some graphics
            >>> execute("linemerge linesimplify linesort write output.svg", doc)

    Args:
        pipeline: vpype pipeline as would be used with ``vpype`` CLI
        document: if provided, is perloaded in the pipeline before the first command executes

    Returns:
        pipeline's content after the last command executes
    """

    if document:

        @cli.command()
        @vp.global_processor
        def vsketchinput(doc):
            doc.extend(document)
            return doc

    out_doc = vp.Document()

    @cli.command()
    @vp.global_processor
    def vsketchoutput(doc):
        out_doc.extend(doc)
        return doc

    args = ("vsketchinput " if document else "") + pipeline + " vsketchoutput"
    cli.main(prog_name="vpype", args=shlex.split(args), standalone_mode=False)
    return out_doc



import os

from PySide2.QtCore import QCoreApplication
from PySide2.QtGui import QIcon, QPalette
from PySide2.QtWidgets import QAction, QActionGroup


def load_icon(path: str) -> QIcon:
    path = os.path.abspath(
        os.path.dirname(__file__) + os.path.sep + "resources" + os.path.sep + path
    )

    # check if dark mode is enabled
    base_color = QCoreApplication.instance().palette().color(QPalette.Base)
    if base_color.lightnessF() < 0.5:
        file, ext = os.path.splitext(path)
        path = file + "-dark" + ext

    icon = QIcon(path)
    return icon


class PenWidthActionGroup(QActionGroup):
    widths = [
        0.05,
        0.1,
        0.15,
        0.2,
        0.25,
        0.3,
        0.35,
        0.4,
        0.45,
        0.5,
        0.6,
        0.7,
        0.8,
        0.9,
        1.0,
        1.2,
    ]

    def __init__(self, current: float = 0.3, parent=None):
        super().__init__(parent)

        if current not in self.widths:
            self.widths.append(current)
        for w in sorted(self.widths):
            act = self.addAction(QAction(f"{w} mm"))
            act.setCheckable(True)
            act.setChecked(w == current)
            act.setData(w)


class PenOpacityActionGroup(QActionGroup):
    opacities = [
        0.3,
        0.5,
        0.7,
        0.8,
        0.9,
        0.95,
        1,
    ]

    def __init__(self, current: float = 0.8, parent=None):
        super().__init__(parent)

        if current not in self.opacities:
            self.opacities.append(current)
        for w in sorted(self.opacities):
            act = self.addAction(QAction(f"{int(w*100)}%"))
            act.setCheckable(True)
            act.setChecked(w == current)
            act.setData(w)



from contextlib import contextmanager
from typing import List, Optional, Union

import click

from .model import Document

# REMINDER: anything added here must be added to docs/api.rst
__all__ = ["VpypeState", "multiple_to_layer_ids", "single_to_layer_id", "LayerType"]


class VpypeState:
    current_state: Union["VpypeState", None] = None

    def __init__(self, doc: Union[Document, None] = None):
        if doc is not None:
            self.document = doc
        else:
            self.document = Document()

        self.target_layer: Optional[int] = None

    @classmethod
    def get_current(cls):
        return cls.current_state

    @contextmanager
    def current(self):
        self.__class__.current_state = self
        yield
        self.__class__.current_state = None


def multiple_to_layer_ids(
    layers: Optional[Union[int, List[int]]], document: Document
) -> List[int]:
    """Convert multiple-layer CLI argument to list of layer IDs.

    Args:
        layers: value from a :class:`LayerType` argument with accept_multiple=True
        document: target :class:`Document` instance

    Returns:
        List of layer IDs
    """
    if layers is None or layers is LayerType.ALL:
        return sorted(document.ids())
    elif isinstance(layers, list):
        return sorted(vid for vid in layers if document.exists(vid))
    else:
        return []


def single_to_layer_id(layer: Optional[int], document: Document) -> int:
    """Convert single-layer CLI argument to layer ID, accounting for the existence of a current
    a current target layer and dealing with default behavior.

    Arg:
        layer: value from a :class:`LayerType` argument
        document: target :class:`Document` instance (for new layer ID)

    Returns:
        Target layer ID
    """
    current_target_layer = VpypeState.get_current().target_layer

    if layer is LayerType.NEW or (layer is None and current_target_layer is None):
        lid = document.free_id()
    elif layer is None:
        lid = VpypeState.get_current().target_layer
    else:
        lid = layer

    return lid


class LayerType(click.ParamType):
    """
    Interpret value of --layer options.

    If `accept_multiple == True`, comma-separated array of int is accepted or 'all'. Returns
    either a list of IDs or `LayerType.ALL`.

    If `accept_new == True`, 'new' is also accepted, in which case returns `LayerType.NEW`.

    None is passed through, which typically means to use the default behaviour.
    """

    NEW = -1
    ALL = -2

    def __init__(self, accept_multiple: bool = False, accept_new: bool = False):
        self.accept_multiple = accept_multiple
        self.accept_new = accept_new

        if accept_multiple:
            self.name = "layers"
        else:
            self.name = "layer"

    def convert(self, value, param, ctx):
        # comply with ParamType requirements
        if value is None:
            return None

        if value.lower() == "all":
            if self.accept_multiple:
                return LayerType.ALL
            else:
                self.fail("'all' was not expected", param, ctx)
        elif value.lower() == "new":
            if self.accept_new:
                return LayerType.NEW
            else:
                self.fail("'new' was not expected", param, ctx)

        try:
            if self.accept_multiple:
                id_arr = list(map(int, value.split(",")))
                for i in id_arr:
                    if i < 1:
                        raise TypeError
                return id_arr
            else:
                return int(value)
        except TypeError:
            self.fail(f"unexpected {value!r} of type {type(value).__name__}", param, ctx)
        except ValueError:
            self.fail(f"{value!r} is not a valid value", param, ctx)



import os
import pathlib
from typing import Optional, Tuple

import cachetools
import cachetools.keys
import moderngl as mgl
import numpy as np
from PIL import Image

ColorType = Tuple[float, float, float, float]


@cachetools.cached(
    cache={},  # type: ignore
    key=lambda name, ctx: cachetools.keys.hashkey((name, id(ctx))),
)
def load_program(name: str, ctx: mgl.Context) -> mgl.Program:
    """Create a program based on a shader name prefix.

    This function automatically load all shader type based on file existence, looking for
    _vertex, _fragment and _geometry suffices in the ``shaders`` directory.

    The same program instance is returned for identical combination of ``name``/``_ctx``.

    Args:
         name: shader file name prefix
         ctx: context

    Returns:
        the loaded program
    """

    def _load_shader(path: str) -> Optional[str]:
        try:
            with open(path) as fp:
                return fp.read()
        except IOError:
            return None

    full_path = os.path.dirname(__file__) + os.path.sep + "shaders" + os.path.sep + name

    return ctx.program(
        vertex_shader=_load_shader(full_path + "_vertex.glsl"),
        fragment_shader=_load_shader(full_path + "_fragment.glsl"),
        geometry_shader=_load_shader(full_path + "_geometry.glsl"),
    )


@cachetools.cached(
    cache={},  # type: ignore
    key=lambda name, ctx, size, components: cachetools.keys.hashkey(
        (name, id(ctx), size, components)
    ),
)
def load_texture_array(
    name: str, ctx: mgl.Context, size: Tuple[int, int, int], components: int = 4
) -> mgl.TextureArray:
    texture_path = pathlib.Path(__file__).parent / "resources" / name
    img = Image.open(str(texture_path))
    texture = ctx.texture_array(size, components, data=img.convert("RGBA").tobytes())
    texture.build_mipmaps()
    return texture


def orthogonal_projection_matrix(
    left: float, right: float, bottom: float, top: float, near: float, far: float, dtype=None
) -> np.ndarray:
    """Creates an orthogonal projection matrix."""

    rml = right - left
    tmb = top - bottom
    fmn = far - near

    a = 2.0 / rml
    b = 2.0 / tmb
    c = -2.0 / fmn
    tx = -(right + left) / rml
    ty = -(top + bottom) / tmb
    tz = -(far + near) / fmn

    # GLSL is column major, thus the transpose
    return np.array(
        (
            (a, 0.0, 0.0, 0.0),
            (0.0, b, 0.0, 0.0),
            (0.0, 0.0, c, 0.0),
            (tx, ty, tz, 1.0),
        ),
        dtype=dtype,
    )



import math
from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Union

import moderngl as mgl
import numpy as np

import vpype as vp

from ._utils import ColorType, load_program, load_texture_array

if TYPE_CHECKING:  # pragma: no cover
    from .engine import Engine

ResourceType = Union[mgl.Buffer, mgl.Texture, mgl.TextureArray]


class Painter:
    def __init__(self, ctx: mgl.Context):
        self._ctx = ctx
        self._resources: List[ResourceType] = []

    def __del__(self):
        for resource in self._resources:
            resource.release()

    def register_resource(self, resource: ResourceType) -> ResourceType:
        self._resources.append(resource)
        return resource

    def buffer(self, *args: Any, **kwargs: Any) -> mgl.Buffer:
        buffer = self._ctx.buffer(*args, **kwargs)
        self.register_resource(buffer)
        return buffer

    def render(self, engine: "Engine", projection: np.ndarray) -> None:
        raise NotImplementedError


class PaperBoundsPainter(Painter):
    def __init__(
        self,
        ctx: mgl.Context,
        paper_size: Tuple[float, float],
        color: ColorType = (0, 0, 0, 0.45),
        shadow_size: float = 7.0,
    ):
        super().__init__(ctx)

        data = np.array(
            [
                (0, 0),
                (paper_size[0], 0),
                (paper_size[0], paper_size[1]),
                (0, paper_size[1]),
                (paper_size[0], shadow_size),
                (paper_size[0] + shadow_size, shadow_size),
                (paper_size[0] + shadow_size, paper_size[1] + shadow_size),
                (shadow_size, paper_size[1] + shadow_size),
                (shadow_size, paper_size[1]),
            ],
            dtype="f4",
        )
        line_idx = np.array([0, 1, 2, 3], dtype="i4")
        triangle_idx = np.array(
            [
                (0, 3, 1),  # page background
                (1, 3, 2),
                (4, 2, 5),  # shadow
                (2, 6, 5),
                (7, 6, 2),
                (8, 7, 2),
            ],
            dtype="i4",
        ).reshape(-1)

        self._color = color
        self._prog = load_program("fast_line_mono", ctx)

        vbo = self.buffer(data.tobytes())
        self._bounds_vao = ctx.vertex_array(
            self._prog, [(vbo, "2f", "in_vert")], self.buffer(line_idx.tobytes())
        )
        self._shading_vao = ctx.vertex_array(
            self._prog, [(vbo, "2f", "in_vert")], self.buffer(triangle_idx.tobytes())
        )

    def render(self, engine: "Engine", projection: np.ndarray) -> None:
        self._prog["projection"].write(projection)

        self._prog["color"].value = (0, 0, 0, 0.25)
        self._shading_vao.render(mgl.TRIANGLES, first=6, vertices=12)

        self._prog["color"].value = (1, 1, 1, 1)
        self._shading_vao.render(mgl.TRIANGLES, first=0, vertices=6)

        self._prog["color"].value = self._color
        self._bounds_vao.render(mgl.LINE_LOOP)


class LineCollectionFastPainter(Painter):
    def __init__(self, ctx: mgl.Context, lc: vp.LineCollection, color: ColorType):
        super().__init__(ctx)

        self._prog = load_program("fast_line_mono", ctx)
        self._color = color

        vertices, indices = self._build_buffers(lc)
        vbo = self.buffer(vertices.tobytes())
        ibo = self.buffer(indices.tobytes())
        self._vao = ctx.vertex_array(self._prog, [(vbo, "2f4", "in_vert")], index_buffer=ibo)

    def render(self, engine: "Engine", projection: np.ndarray) -> None:
        self._prog["projection"].write(projection)
        self._prog["color"].value = self._color
        self._vao.render(mgl.LINE_STRIP)

    @staticmethod
    def _build_buffers(lc: vp.LineCollection) -> Tuple[np.ndarray, np.ndarray]:
        total_length = sum(len(line) for line in lc)
        buffer = np.empty((total_length, 2), dtype="f4")
        indices = np.empty(total_length + len(lc), dtype="i4")
        indices.fill(-1)

        # build index array
        cur_index = 0
        for i, line in enumerate(lc):
            next_idx = cur_index + len(line)
            indices[i + cur_index : i + next_idx] = np.arange(cur_index, next_idx)
            buffer[cur_index:next_idx] = vp.as_vector(line)
            cur_index = next_idx

        return buffer, indices


class LineCollectionFastColorfulPainter(Painter):
    COLORS = [
        np.array((0.0, 0.0, 1.0, 1.0)),
        np.array((0.0, 0.5, 0.0, 1.0)),
        np.array((1.0, 0.0, 0.0, 1.0)),
        np.array((0.0, 0.75, 0.75, 1.0)),
        np.array((0.0, 1.0, 0.0, 1.0)),
        np.array((0.75, 0, 0.75, 1.0)),
        np.array((0.75, 0.75, 0.0, 1.0)),
    ]

    def __init__(self, ctx: mgl.Context, lc: vp.LineCollection, show_points: bool = False):
        super().__init__(ctx)

        self._show_points = show_points
        self._prog = load_program("fast_line", ctx)

        # TODO: hacked color table size is not ideal, this will need to be changed when
        # implementing color themes
        self._prog["colors"].write(np.concatenate(self.COLORS).astype("f4").tobytes())

        vertices, indices = self._build_buffers(lc)
        vbo = self.buffer(vertices.tobytes())
        ibo = self.buffer(indices.tobytes())

        self._vao = ctx.vertex_array(
            self._prog,
            [(vbo, "2f4 i1", "in_vert", "color_idx")],
            ibo,
        )

    def render(self, engine: "Engine", projection: np.ndarray) -> None:
        self._prog["projection"].write(projection)
        self._vao.render(mgl.LINE_STRIP)
        if self._show_points:
            self._vao.render(mgl.POINTS)

    @classmethod
    def _build_buffers(cls, lc: vp.LineCollection) -> Tuple[np.ndarray, np.ndarray]:
        total_length = sum(len(line) for line in lc)
        buffer = np.empty(total_length, dtype=[("vertex", "2f4"), ("color", "i1")])
        indices = np.empty(total_length + len(lc), dtype="i4")
        indices.fill(-1)

        # build index array
        cur_index = 0
        for i, line in enumerate(lc):
            next_idx = cur_index + len(line)
            indices[i + cur_index : i + next_idx] = np.arange(cur_index, next_idx)

            buffer["vertex"][cur_index:next_idx] = vp.as_vector(line)
            buffer["color"][cur_index:next_idx] = i % len(cls.COLORS)

            cur_index = next_idx

        return buffer, indices


class LineCollectionPointsPainter(Painter):
    def __init__(
        self, ctx: mgl.Context, lc: vp.LineCollection, color: ColorType = (0, 0, 0, 0.25)
    ):
        super().__init__(ctx)

        vertex = """
            #version 330
            
            uniform mat4 projection;
            
            in vec2 position;
            
            void main() {
              gl_PointSize = 5.0;
              gl_Position = projection * vec4(position, 0.0, 1.0);
            }
        """

        fragment = """
            #version 330
            
            uniform vec4 color;
            
            out vec4 out_color;
            
            void main() {
               out_color = color;
            }
        """

        self._prog = ctx.program(vertex_shader=vertex, fragment_shader=fragment)
        self._color = color

        vertices = self._build_buffers(lc)
        vbo = self.buffer(vertices.tobytes())
        self._vao = ctx.vertex_array(self._prog, [(vbo, "2f4", "position")])

    def render(self, engine: "Engine", projection: np.ndarray) -> None:
        self._prog["projection"].write(projection)
        self._prog["color"].value = self._color
        self._vao.render(mgl.POINTS)

    @staticmethod
    def _build_buffers(lc: vp.LineCollection) -> np.ndarray:
        buffer = np.empty((sum(len(line) for line in lc), 2), dtype="f4")

        # build index array
        cur_index = 0
        for i, line in enumerate(lc):
            next_idx = cur_index + len(line)
            buffer[cur_index:next_idx] = vp.as_vector(line)
            cur_index = next_idx

        return buffer


class LineCollectionPenUpPainter(Painter):
    def __init__(
        self, ctx: mgl.Context, lc: vp.LineCollection, color: ColorType = (0, 0, 0, 0.5)
    ):
        super().__init__(ctx)

        self._color = color
        self._prog = load_program("fast_line_mono", ctx)

        # build vertices
        vertices: List[Tuple[float, float]] = []
        for i in range(len(lc) - 1):
            vertices.extend(
                ((lc[i][-1].real, lc[i][-1].imag), (lc[i + 1][0].real, lc[i + 1][0].imag))
            )

        if len(vertices) > 0:
            vbo = self.buffer(np.array(vertices, dtype="f4").tobytes())
            self._vao = ctx.vertex_array(self._prog, [(vbo, "2f4", "in_vert")])
        else:
            self._vao = None

    def render(self, engine: "Engine", projection: np.ndarray) -> None:
        if self._vao is not None:
            self._prog["color"].value = self._color
            self._prog["projection"].write(projection)
            self._vao.render(mgl.LINES)


class LineCollectionPreviewPainter(Painter):
    def __init__(
        self, ctx: mgl.Context, lc: vp.LineCollection, pen_width: float, color: ColorType
    ):
        super().__init__(ctx)

        self._color = color
        self._pen_width = pen_width
        self._prog = load_program("preview_line", ctx)

        vertices, indices = self._build_buffers(lc)
        vbo = self.buffer(vertices.tobytes())
        ibo = self.buffer(indices.tobytes())
        self._vao = ctx.vertex_array(self._prog, [(vbo, "2f4", "position")], ibo)

    def render(self, engine: "Engine", projection: np.ndarray) -> None:
        self._prog["color"].value = self._color
        self._prog["pen_width"].value = self._pen_width
        self._prog["antialias"].value = 1.5 / engine.scale
        self._prog["projection"].write(projection)

        if engine.debug:
            self._prog["kill_frag_shader"].value = False
            self._prog["debug_view"].value = True
            self._prog["color"].value = self._color[0:3] + (0.3,)
            self._vao.render(mgl.LINE_STRIP_ADJACENCY)

            self._prog["kill_frag_shader"].value = True
            self._prog["debug_view"].value = False
            self._prog["color"].value = (0, 1, 0, 1)
            self._ctx.wireframe = True
            self._vao.render(mgl.LINE_STRIP_ADJACENCY)
            self._ctx.wireframe = False
        else:
            self._prog["kill_frag_shader"].value = False
            self._prog["debug_view"].value = False
            self._vao.render(mgl.LINE_STRIP_ADJACENCY)

    @staticmethod
    def _build_buffers(lc: vp.LineCollection):
        """Prepare the buffers for multi-polyline rendering. Closed polyline must have their
        last point identical to their first point."""

        indices = []
        reset_index = [-1]
        start_index = 0
        for i, line in enumerate(lc):
            if line[0] == line[-1]:  # closed path
                idx = np.arange(len(line) + 3) - 1
                idx[0], idx[-2], idx[-1] = len(line) - 1, 0, 1
            else:
                idx = np.arange(len(line) + 2) - 1
                idx[0], idx[-1] = 0, len(line) - 1

            indices.append(idx + start_index)
            start_index += len(line)
            indices.append(reset_index)

        return (
            np.vstack([vp.as_vector(line).astype("f4") for line in lc]),
            np.concatenate(indices).astype("i4"),
        )


class RulersPainter(Painter):
    def __init__(self, ctx: mgl.Context):
        super().__init__(ctx)

        # this also sets the font size
        self._thickness = 20.0
        self._font_size = 7.0

        self._prog = load_program("ruler_patch", ctx)

        # vertices
        vertices = self.buffer(
            np.array(
                [
                    (-1.0, 1.0),
                    (0.0, 1.0),
                    (1.0, 1.0),
                    (-1.0, 0.0),
                    (0, 0.0),
                    (1.0, 0.0),
                    (-1.0, -1.0),
                    (0.0, -1.0),
                ],
                dtype="f4",
            ).tobytes()
        )

        # line strip for stroke
        frame_indices = self.buffer(np.array([3, 5, 1, 7], dtype="i4").tobytes())
        self._stroke_vao = ctx.vertex_array(
            self._prog, [(vertices, "2f4", "in_vert")], frame_indices
        )

        # triangles for fill
        # first 6 vertices for the small top-right
        # next 12 vertices for the rulers themselves
        patch_indices = self.buffer(
            np.array(
                [0, 1, 3, 1, 3, 4, 1, 2, 4, 2, 4, 5, 3, 4, 6, 4, 6, 7], dtype="i4"
            ).tobytes()
        )
        self._fill_vao = ctx.vertex_array(
            self._prog, [(vertices, "2f4", "in_vert")], patch_indices
        )

        # major ticks buffer
        self._ticks_prog = load_program("ruler_ticks", ctx)
        self._ticks_prog["color"] = (0.2, 0.2, 0.2, 1.0)
        self._ticks_vao = ctx.vertex_array(self._ticks_prog, [])

        # TEXT STUFF
        # https://github.com/Contraz/demosys-py/blob/master/demosys/effects/text/resources/data/demosys/text/meta.json
        # {
        #     "characters": 190,
        #     "character_ranges": [
        #         {
        #             "min": 32,
        #             "max": 126
        #         },
        #         {
        #             "min": 161,
        #             "max": 255
        #         }
        #     ],
        #     "character_height": 159,
        #     "character_width": 77,
        #     "atlas_height": 30210,
        #     "atlas_width": 77
        # }
        self._texture = load_texture_array("VeraMono.png", ctx, (77, 159, 190), 4)
        self._text_prog = load_program("ruler_text", ctx)
        self._aspect_ratio = 159.0 / 77.0
        self._text_prog["color"].value = (0, 0, 0, 1.0)
        self._text_vao = ctx.vertex_array(self._text_prog, [])

        # unit label
        self._unit_label = LabelPainter(ctx, "XX")

    @property
    def thickness(self) -> float:
        return self._thickness

    def render(self, engine: "Engine", projection: np.ndarray) -> None:
        # ===========================
        # render frame

        self._prog["ruler_width"] = 2 * self._thickness * engine.pixel_factor / engine.width
        self._prog["ruler_height"] = 2 * self._thickness * engine.pixel_factor / engine.height
        self._prog["color"].value = (1.0, 1.0, 1.0, 1.0)
        self._fill_vao.render(mode=mgl.TRIANGLES, first=6)

        # ===========================
        # render ticks

        spec = engine.scale_spec

        self._ticks_prog["scale"] = spec.scale_px * engine.scale
        self._ticks_prog["divisions"] = list(spec.divisions)
        self._ticks_prog["delta_number"] = spec.scale

        # compute various stuff
        horiz_tick_count = math.ceil(engine.width / engine.scale / spec.scale_px) + 1
        vertical_tick_count = math.ceil(engine.height / engine.scale / spec.scale_px) + 1
        doc_width, doc_height = (
            engine.document.page_size
            if engine.document is not None and engine.document.page_size is not None
            else (-1.0, -1.0)
        )
        start_number_horiz = math.floor(engine.origin[0] / spec.scale_px) * spec.scale
        start_number_vert = math.floor(engine.origin[1] / spec.scale_px) * spec.scale
        thickness = self._thickness * engine.pixel_factor
        font_size = self._font_size * engine.pixel_factor

        # render vertical ruler
        self._ticks_prog["vertical"] = True
        self._ticks_prog["viewport_dim"] = engine.height
        self._ticks_prog["document_dim"] = doc_height / spec.to_px
        self._ticks_prog["offset"] = (engine.origin[1] % spec.scale_px) * engine.scale
        self._ticks_prog["ruler_thickness"] = 2 * thickness / engine.width
        self._ticks_prog["start_number"] = start_number_vert
        self._ticks_vao.render(mode=mgl.POINTS, vertices=vertical_tick_count)

        # render horizontal ruler
        self._ticks_prog["vertical"] = False
        self._ticks_prog["viewport_dim"] = engine.width
        self._ticks_prog["document_dim"] = doc_width / spec.to_px
        self._ticks_prog["offset"] = (engine.origin[0] % spec.scale_px) * engine.scale
        self._ticks_prog["ruler_thickness"] = 2 * thickness / engine.height
        self._ticks_prog["start_number"] = start_number_horiz
        self._ticks_vao.render(mode=mgl.POINTS, vertices=horiz_tick_count)

        # ===========================
        # render glyph
        self._texture.use(0)
        self._text_prog["scale"] = spec.scale_px * engine.scale
        self._text_prog["delta_number"] = spec.scale

        # horizontal
        self._text_prog["vertical"] = False
        self._text_prog["viewport_dim"] = engine.width
        self._text_prog["document_dim"] = doc_width / spec.to_px
        self._text_prog["offset"] = (engine.origin[0] % spec.scale_px) * engine.scale
        self._text_prog["glyph_size"].value = (
            font_size * 2.0 / engine.width,
            font_size * 2.0 * self._aspect_ratio / engine.height,
        )
        self._text_prog["start_number"] = start_number_horiz
        self._text_vao.render(mode=mgl.POINTS, vertices=horiz_tick_count)

        # vertical
        self._text_prog["vertical"] = True
        self._text_prog["viewport_dim"] = engine.height
        self._text_prog["document_dim"] = doc_height / spec.to_px
        self._text_prog["offset"] = (engine.origin[1] % spec.scale_px) * engine.scale
        self._text_prog["glyph_size"].value = (
            font_size * 2.0 * self._aspect_ratio / engine.width,
            font_size * 2.0 / engine.height,
        )
        self._text_prog["start_number"] = start_number_vert
        self._text_vao.render(mode=mgl.POINTS, vertices=vertical_tick_count)

        # ===========================
        # render units corner

        self._prog["color"].value = (1.0, 1.0, 1.0, 1.0)
        self._fill_vao.render(mode=mgl.TRIANGLES, vertices=6)
        self._prog["color"].value = (0.0, 0.0, 0.0, 1.0)
        self._stroke_vao.render(mode=mgl.LINES)

        self._unit_label.font_size = font_size
        self._unit_label.position = (thickness / 7.0, thickness / 8.0)
        self._unit_label.label = spec.unit
        self._unit_label.render(engine, projection)


class LabelPainter(Painter):
    def __init__(
        self,
        ctx: mgl.Context,
        label: str = "",
        position: Tuple[float, float] = (0.0, 0.0),
        font_size: float = 14.0,
        max_size: Optional[int] = None,
        color: ColorType = (0.0, 0.0, 0.0, 1.0),
    ):
        super().__init__(ctx)

        self.position = position
        self.font_size = font_size
        self._max_size = max_size or len(label)
        self._buffer = self.buffer(reserve=self._max_size)
        self.label = label
        self._color = color

        self._texture = load_texture_array("VeraMono.png", ctx, (77, 159, 190), 4)
        self._aspect_ratio = 159.0 / 77.0
        self._prog = load_program("label", ctx)
        self._vao = ctx.vertex_array(self._prog, [(self._buffer, "u1", "char")])

    @property
    def label(self) -> str:
        return self._label

    @label.setter
    def label(self, label: str) -> None:
        self._label = label
        self._size = min(len(label), self._max_size)
        self._buffer.write(
            np.array([ord(c) for c in label[: self._max_size]], dtype=np.uint8).tobytes()
        )

    def render(self, engine: "Engine", projection: np.ndarray) -> None:
        self._texture.use(0)
        self._prog["color"].value = self._color
        self._prog["position"].value = (
            -1.0 + 2.0 * self.position[0] / engine.width,
            1.0 - 2.0 * self.position[1] / engine.height,
        )
        self._prog["glyph_size"].value = (
            self.font_size * 2.0 / engine.width,
            self.font_size * 2.0 * self._aspect_ratio / engine.height,
        )

        self._vao.render(mode=mgl.POINTS, vertices=self._size)



"""
Hidden debug commands to help testing.
"""
import json
from typing import Any, Dict, Iterable, List, Sequence

import numpy as np

from vpype import Document, LineCollection, as_vector, global_processor

from .cli import cli

debug_data: List[Dict[str, Any]] = []

__all__ = ("dbsample", "dbdump", "stat", "DebugData")


@cli.command(hidden=True)
@global_processor
def dbsample(document: Document):
    """
    Show statistics on the current geometries in JSON format.
    """
    global debug_data

    data: Dict[str, Any] = {}
    if document.is_empty():
        data["count"] = 0
    else:
        data["count"] = sum(len(lc) for lc in document.layers.values())
        data["layer_count"] = len(document.layers)
        data["length"] = document.length()
        data["pen_up_length"] = document.pen_up_length()
        data["bounds"] = document.bounds()
        data["layers"] = {
            layer_id: [as_vector(line).tolist() for line in layer]
            for layer_id, layer in document.layers.items()
        }

    debug_data.append(data)
    return document


@cli.command(hidden=True)
@global_processor
def dbdump(document: Document):
    global debug_data
    print(json.dumps(debug_data))
    debug_data = []
    return document


class DebugData:
    """
    Helper class to load
    """

    @staticmethod
    def load(debug_output: str):
        """
        Create DebugData instance array from debug output
        :param debug_output:
        :return:
        """
        return [DebugData(data) for data in json.loads(debug_output)]

    def __init__(self, data: Dict[str, Any]):
        self.count = data["count"]
        self.length = data.get("length", 0)
        self.pen_up_length = data.get("pen_up_length", 0)
        self.bounds = data.get("bounds", [0, 0, 0, 0])
        self.layers = data.get("layers", {})

        self.document = Document()
        for vid, lines in self.layers.items():
            self.document[int(vid)] = LineCollection(
                [np.array([x + 1j * y for x, y in line]) for line in lines]
            )

    def bounds_within(self, x: float, y: float, width: float, height: float) -> bool:
        """
        Test if coordinates are inside. If `x` and `y` are provided only, consider input as
        a point. If `width` and `height` are passed as well, consider input as rect.
        """
        if self.count == 0:
            return False

        def approx_check(a, b, lt):
            if lt:
                return a < b and not np.isclose(a, b)
            else:
                return a > b and not np.isclose(a, b)

        if (
            approx_check(self.bounds[0], x, True)
            or approx_check(self.bounds[1], y, True)
            or approx_check(self.bounds[2], x + width, False)
            or approx_check(self.bounds[3], y + height, False)
        ):
            return False

        return True

    def __eq__(self, other: object):
        if not isinstance(other, DebugData):
            return NotImplemented

        if self.count == 0:
            return other.count == 0

        return (
            np.all(np.isclose(np.array(self.bounds), np.array(other.bounds)))
            and self.length == other.length
        )

    def has_layer(self, lid: int) -> bool:
        return self.has_layers([lid])

    def has_layers(self, lids: Iterable[int]) -> bool:
        return all(str(lid) in self.layers for lid in lids)

    def has_layer_only(self, lid: int) -> bool:
        return self.has_layers_only([lid])

    def has_layers_only(self, lids: Sequence[int]) -> bool:
        return self.has_layers(lids) and len(self.layers.keys()) == len(lids)


@cli.command(group="Output")
@global_processor
def stat(document: Document):
    """Print human-readable statistics on the current geometries."""

    print("========= Stats ========= ")
    print(f"Current page size: {document.page_size}")
    length_tot = 0.0
    pen_up_length_tot = 0.0
    for layer_id in sorted(document.layers.keys()):
        layer = document.layers[layer_id]
        length = layer.length()
        pen_up_length, pen_up_mean, pen_up_median = layer.pen_up_length()
        length_tot += length
        pen_up_length_tot += pen_up_length
        print(f"Layer {layer_id}")
        print(f"  Length: {length}")
        print(f"  Pen-up length: {pen_up_length}")
        print(f"  Total length: {length + pen_up_length}")
        print(f"  Mean pen-up length: {pen_up_mean}")
        print(f"  Median pen-up length: {pen_up_median}")
        print(f"  Path count: {len(layer)}")
        print(f"  Segment count: {layer.segment_count()}")
        print(
            f"  Mean segment length:",
            str(length / layer.segment_count() if layer.segment_count() else "n/a"),
        )
        print(f"  Bounds: {layer.bounds()}")
    print(f"Totals")
    print(f"  Layer count: {len(document.layers)}")
    print(f"  Length: {length_tot}")
    print(f"  Pen-up length: {pen_up_length_tot}")
    print(f"  Total length: {length_tot + pen_up_length_tot}")
    print(f"  Path count: {sum(len(layer) for layer in document.layers.values())}")
    print(f"  Segment count: {document.segment_count()}")

    print(
        f"  Mean segment length:",
        str(length_tot / document.segment_count() if document.segment_count() else "n/a"),
    )
    print(f"  Bounds: {document.bounds()}")
    print("========================= ")

    return document



"""Config file support for vpype.

Configuration data is accessed via the ``config_manager`` global variable::

    >>> from vpype import config_manager
    >>> config_manager.config  # dictionary of all configuration

HPGL plotter have specific config support::

    >>> plotter_config = config_manager.get_plotter_config("hp7475a")
    >>> plotter_config
    PlotterConfig(name='hp7475a', ...)
    >>> plotter_config.paper_config("a4")
    PaperConfig(name='a4', paper_size=(1122.5196850393702, 793.7007874015749), ...)
"""

import logging
import math
import os
import pathlib
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple, Union

import attr
import toml

from .utils import convert_length

__all__ = [
    "PaperConfig",
    "PlotterConfig",
    "ConfigManager",
    "config_manager",
    "CONFIG_MANAGER",
]


def _convert_length_pair(data: Sequence[Union[float, str]]) -> Tuple[float, float]:
    return convert_length(data[0]), convert_length(data[1])


@attr.s(auto_attribs=True, frozen=True)
class PaperConfig:
    """Data class containing configuration for a give plotter type/paper size combinations."""

    name: str  #: name of the paper format
    y_axis_up: bool  #: if True, the Y axis point upwards instead of downwards
    origin_location: Tuple[
        float, float
    ]  #: location on paper of the (0, 0) plotter unit coordinates

    paper_size: Optional[Tuple[float, float]] = None  #: X/Y axis convention of the plotter
    paper_orientation: Optional[
        str
    ] = None  #: orientation of the plotter coordinate system on paper
    x_range: Optional[Tuple[int, int]] = None  #: admissible range of X coordinates
    y_range: Optional[Tuple[int, int]] = None  #: admissible range of Y coordinates
    origin_location_reference: Optional[str] = "topleft"  #: reference for ``origin_location``

    info: str = ""  #: information printed to the user when paper is used
    rotate_180: bool = False  #: if True, the geometries are rotated by 180 degrees on the page
    set_ps: Optional[int] = None  #: if not None, call PS with corresponding value
    final_pu_params: Optional[
        str
    ] = None  #: if not None, these params are added to the final PU command
    aka_names: List[
        str
    ] = []  #: alternative paper names (will be found by :func:`paper_config`

    @classmethod
    def from_config(cls, data: Dict[str, Any]) -> "PaperConfig":
        return cls(
            name=data["name"],
            y_axis_up=data["y_axis_up"],
            origin_location=_convert_length_pair(data["origin_location"]),
            paper_size=_convert_length_pair(data["paper_size"])
            if "paper_size" in data
            else None,
            paper_orientation=data.get("paper_orientation", None),
            x_range=(data["x_range"][0], data["x_range"][1]) if "x_range" in data else None,
            y_range=(data["y_range"][0], data["y_range"][1]) if "y_range" in data else None,
            origin_location_reference=data.get("origin_location_reference", "topleft"),
            info=data.get("info", ""),
            rotate_180=data.get("rotate_180", False),
            set_ps=data.get("set_ps", None),
            final_pu_params=data.get("final_pu_params", None),
            aka_names=data.get("aka_names", []),
        )


@attr.s(auto_attribs=True, frozen=True)
class PlotterConfig:
    """Data class containing configuration for a given plotter type."""

    name: str  #: name of the plotter
    paper_configs: List[PaperConfig]  #: list of :class:`PaperConfig` instance
    plotter_unit_length: float  #: physical size of plotter units (in pixel)
    pen_count: int  #: number of pen supported by the plotter

    info: str = ""  #: information printed to the user when plotter is used

    @classmethod
    def from_config(cls, data: Dict[str, Any]) -> "PlotterConfig":
        return cls(
            name=data["name"],
            paper_configs=[PaperConfig.from_config(d) for d in data["paper"]],
            plotter_unit_length=convert_length(data["plotter_unit_length"]),
            pen_count=data["pen_count"],
            info=data.get("info", ""),
        )

    def paper_config(self, paper: str) -> Optional[PaperConfig]:
        """Return the paper configuration for ``paper`` or none if not found.

        Args:
            paper: desired paper format designator

        Returns:
            the :class:`PaperConfig` instance corresponding to ``paper`` or None if not found
        """
        for pc in self.paper_configs:
            if paper == pc.name or paper in pc.aka_names:
                return pc
        return None

    def paper_config_from_size(
        self, page_size: Optional[Tuple[float, float]]
    ) -> Optional[PaperConfig]:
        """Look for a paper configuration matching ``paper_format`` and return it if found.

        Args:
            page_size: tuple of desired page size (may be ``None``, in which case ``None`` is
                returned

        Returns:
            the :class:`PaperConfig` instance corresponding to ``paper_format`` or None if not
            found
        """

        if page_size is None:
            return None

        def _isclose_tuple(a, b):
            return all(math.isclose(aa, bb) for aa, bb in zip(a, b))

        for pc in self.paper_configs:
            if _isclose_tuple(pc.paper_size, page_size) or _isclose_tuple(
                pc.paper_size, tuple(reversed(page_size))
            ):
                return pc
        return None


class ConfigManager:
    """Helper class to handle vpype's TOML configuration files.

    This class is typically used via its singleton instance ``config_manager``::

        >>> from vpype import config_manager
        >>> my_config = config_manager.config.get("my_config", None)

    Helper methods are provided for specific aspects of configuration, such as command-specific
    configs or HPGL-related configs.

    By default, built-in configuration packaged with vpype are loaded at startup. If a file
    exists at path ``~/.vpype.toml``, it will be loaded as well. Additionaly files may be
    loaded using the :func:`load_config_file` method.
    """

    def __init__(self):
        self._config: Dict = {}

    def load_config_file(self, path: str) -> None:
        """Load a config file and add its content to the configuration database. The
        configuration file must be in TOML format.

        Args:
            path: path of the config file
        """

        def _update(d: Dict, u: Mapping) -> Dict:
            """This function must overwrite list member, UNLESS they are list of table, in
            which case they must extend the list."""
            for k, v in u.items():
                if isinstance(v, dict):
                    d[k] = _update(d.get(k, {}), v)
                elif isinstance(v, list) and len(v) > 0 and isinstance(v[0], dict):
                    if k in d:
                        d[k].extend(v)
                    else:
                        d[k] = v
                else:
                    d[k] = v
            return d

        logging.info(f"loading config file at {path}")
        self._config = _update(self._config, toml.load(path))

    def get_plotter_list(self) -> List[str]:
        """Returns a list of plotter names whose configuration is available.

        Returns:
            list of plotter name
        """
        return list(self.config.get("device", {}).keys())

    def get_plotter_config(self, name: Optional[str]) -> Optional[PlotterConfig]:
        """Returns a :class:`PlotterConfig` instance for plotter ``name``.

        Args:
            name: name of desired plotter (may be ``None``, in which case ``None`` is returned)

        Returns:
            :class:`PlotterConfig` instance or None if not found
        """
        devices = self.config.get("device", {})
        if name in devices:
            return PlotterConfig.from_config(devices[name])
        else:
            return None

    def get_command_config(self, name: str) -> Dict[str, Any]:
        """Returns the configuration for command ``name``.

        Args:
            name: name of the command

        Returns:
            dictionary containing the config values (empty if none defined)
        """
        commands = self.config.get("command", {})
        if name in commands:
            return commands[name]
        else:
            return {}

    @property
    def config(self) -> Dict[str, Any]:
        return self._config


config_manager = ConfigManager()

# deprecated
CONFIG_MANAGER = config_manager


def _init():
    config_manager.load_config_file(str(pathlib.Path(__file__).parent / "hpgl_devices.toml"))
    path = os.path.expanduser("~/.vpype.toml")
    if os.path.exists(path):
        config_manager.load_config_file(str(path))


_init()



import logging
from typing import List, Optional, Tuple, Union, cast

import click
import numpy as np

import vpype as vp

from .cli import cli

__all__ = (
    "crop",
    "filter_command",
    "layout",
    "linemerge",
    "linesimplify",
    "linesort",
    "multipass",
    "pagesize",
    "reloop",
    "snap",
    "splitall",
    "trim",
)


@cli.command(group="Operations")
@click.argument("x", type=vp.LengthType(), required=True)
@click.argument("y", type=vp.LengthType(), required=True)
@click.argument("width", type=vp.LengthType(), required=True)
@click.argument("height", type=vp.LengthType(), required=True)
@vp.layer_processor
def crop(lines: vp.LineCollection, x: float, y: float, width: float, height: float):
    """Crop the geometries.

    The crop area is defined by the (X, Y) top-left corner and the WIDTH and HEIGHT arguments.
    All arguments understand supported units.
    """

    lines.crop(x, y, x + width, y + height)
    return lines


@cli.command(group="Operations")
@click.argument("margin_x", type=vp.LengthType(), required=True)
@click.argument("margin_y", type=vp.LengthType(), required=True)
@click.option(
    "-l",
    "--layer",
    type=vp.LayerType(accept_multiple=True),
    default="all",
    help="Target layer(s).",
)
@vp.global_processor
def trim(
    document: vp.Document, margin_x: float, margin_y: float, layer: Union[int, List[int]]
) -> vp.Document:
    """Trim the geometries by some margin.

    This command trims the geometries by the provided X and Y margins with respect to the
    current bounding box.

    By default, `trim` acts on all layers. If one or more layer IDs are provided with the
    `--layer` option, only these layers will be affected. In this case, the bounding box is
    that of the listed layers.
    """

    layer_ids = vp.multiple_to_layer_ids(layer, document)
    bounds = document.bounds(layer_ids)

    if not bounds:
        return document

    min_x = bounds[0] + margin_x
    max_x = bounds[2] - margin_x
    min_y = bounds[1] + margin_y
    max_y = bounds[3] - margin_y
    if min_x > max_x:
        min_x = max_x = 0.5 * (min_x + max_x)
    if min_y > max_y:
        min_y = max_y = 0.5 * (min_y + max_y)

    for vid in layer_ids:
        lc = document[vid]
        lc.crop(min_x, min_y, max_x, max_y)

    return document


@cli.command(group="Operations")
@click.option(
    "-t",
    "--tolerance",
    type=vp.LengthType(),
    default="0.05mm",
    help="Maximum distance between two line endings that should be merged.",
)
@click.option(
    "-f", "--no-flip", is_flag=True, help="Disable reversing stroke direction for merging."
)
@vp.layer_processor
def linemerge(lines: vp.LineCollection, tolerance: float, no_flip: bool = True):
    """
    Merge lines whose endings and starts overlap or are very close.

    By default, `linemerge` considers both directions of a stroke. If there is no additional
    start of a stroke within the provided tolerance, it also checks for ending points of
    strokes and uses them in reverse. You can use the `--no-flip` to disable this reversing
    behaviour and preserve the stroke direction from the input.

    By default, gaps of maximum 0.05mm are considered for merging. This can be controlled with
    the `--tolerance` option.
    """

    lines.merge(tolerance=tolerance, flip=not no_flip)
    return lines


@cli.command(group="Operations")
@click.option(
    "-f",
    "--no-flip",
    is_flag=True,
    help="Disable reversing stroke direction for optimization.",
)
@vp.layer_processor
def linesort(lines: vp.LineCollection, no_flip: bool = True):
    """
    Sort lines to minimize the pen-up travel distance.

    Note: this process can be lengthy depending on the total number of line. Consider using
    `linemerge` before `linesort` to reduce the total number of line and thus significantly
    optimizing the overall plotting time.
    """
    if len(lines) < 2:
        return lines

    index = vp.LineIndex(lines[1:], reverse=not no_flip)
    new_lines = vp.LineCollection([lines[0]])

    while len(index) > 0:
        idx, reverse = index.find_nearest(new_lines[-1][-1])
        line = index.pop(idx)
        if reverse:
            line = np.flip(line)
        new_lines.append(line)

    logging.info(
        f"optimize: reduced pen-up (distance, mean, median) from {lines.pen_up_length()} to "
        f"{new_lines.pen_up_length()}"
    )

    return new_lines


@cli.command(group="Operations")
@click.option(
    "-t",
    "--tolerance",
    type=vp.LengthType(),
    default="0.05mm",
    help="Controls how far from the original geometry simplified points may lie.",
)
@vp.layer_processor
def linesimplify(lines: vp.LineCollection, tolerance):
    """
    Reduce the number of segments in the geometries.

    The resulting geometries' points will be at a maximum distance from the original controlled
    by the `--tolerance` parameter (0.05mm by default).
    """
    if len(lines) < 2:
        return lines

    # Note: preserve_topology must be False, otherwise non-simple (ie intersecting) MLS will
    # not be simplified (see https://github.com/Toblerity/Shapely/issues/911)
    mls = lines.as_mls().simplify(tolerance=tolerance, preserve_topology=False)
    new_lines = vp.LineCollection(mls)

    logging.info(
        f"simplify: reduced segment count from {lines.segment_count()} to "
        f"{new_lines.segment_count()}"
    )

    return new_lines


@cli.command(group="Operations")
@click.option(
    "-t",
    "--tolerance",
    type=vp.LengthType(),
    default="0.05mm",
    help="Controls how close the path beginning and end must be to consider it closed ("
    "default: 0.05mm).",
)
@vp.layer_processor
def reloop(lines: vp.LineCollection, tolerance):
    """Randomize the seam location of closed paths.

    When plotted, closed path may exhibit a visible mark at the seam, i.e. the location where
    the pen begins and ends the stroke. This command randomizes the seam location in order to
    help reduce visual effect of this in plots with regular patterns.

    Paths are considered closed when their beginning and end points are closer than some
    tolerance, which can be set with the `--tolerance` option.
    """

    lines.reloop(tolerance=tolerance)
    return lines


@cli.command(group="Operations")
@click.option(
    "-n", "--count", type=int, default=2, help="How many pass for each line (default: 2)."
)
@vp.layer_processor
def multipass(lines: vp.LineCollection, count: int):
    """
    Add multiple passes to each line

    Each line is extended with a mirrored copy of itself, optionally multiple times. This is
    useful for pens that need several passes to ensure a good quality.
    """
    if count < 2:
        return lines

    new_lines = vp.LineCollection()
    for line in lines:
        new_lines.append(
            np.hstack(
                [line] + [line[-2::-1] if i % 2 == 0 else line[1:] for i in range(count - 1)]
            )
        )

    return new_lines


@cli.command(group="Operations")
@vp.layer_processor
def splitall(lines: vp.LineCollection) -> vp.LineCollection:
    """
    Split all paths into their constituent segments.

    This command may be used together with `linemerge` for cases such as densely-connected
    meshes where the latter cannot optimize well enough by itself. This command will
    filter out segments with identical end-points.

    Note that since some paths (especially curved ones) can be made of a large number of
    segments, this command may significantly increase the processing time of the pipeline.
    """

    new_lines = vp.LineCollection()
    for line in lines:
        new_lines.extend(
            [line[i : i + 2] for i in range(len(line) - 1) if line[i] != line[i + 1]]
        )
    return new_lines


@cli.command(name="filter", group="Operations")
@click.option(
    "--min-length",
    "-m",
    type=vp.LengthType(),
    help="keep lines whose length is no shorter than value",
)
@click.option(
    "--max-length",
    "-M",
    type=vp.LengthType(),
    help="keep lines whose length is no greater than value",
)
@click.option("--closed", "-c", is_flag=True, help="keep closed lines")
@click.option("--not-closed", "-o", is_flag=True, help="reject closed lines")
@click.option(
    "--tolerance",
    "-t",
    type=vp.LengthType(),
    default="0.05mm",
    help="tolerance used to determined if a line is closed or not (default: 0.05mm)",
)
@vp.layer_processor
def filter_command(
    lines: vp.LineCollection,
    min_length: Optional[float],
    max_length: Optional[float],
    closed: bool,
    not_closed: bool,
    tolerance: float,
) -> vp.LineCollection:
    """Filter paths according to specified criterion.

    When an option is provided (e.g. `--min-length 10cm`) the corresponding criterion is
    applied and paths which do not respect the criterion (e.g. a 9cm-long path) are rejected.

    If multiple options are provided, paths will be kept only if they respect every
    corresponding criterion (i.e. logical AND operator).
    """
    keys = []
    if min_length is not None:
        keys.append(lambda line: vp.line_length(line) >= cast(float, min_length))
    if max_length is not None:
        keys.append(lambda line: vp.line_length(line) <= cast(float, max_length))
    if closed:
        keys.append(lambda line: vp.is_closed(line, tolerance))
    if not_closed:
        keys.append(lambda line: not vp.is_closed(line, tolerance))

    if keys:
        lines.filter(lambda line: vp.union(line, keys))
    else:
        logging.warning("filter: no criterion was provided, all geometries are preserved")

    return lines


def _normalize_page_size(
    page_size: Tuple[float, float], landscape: bool
) -> Tuple[float, float]:
    """Normalize page size to respect the orientation."""
    if (landscape and page_size[0] < page_size[1]) or (
        not landscape and page_size[0] > page_size[1]
    ):
        return page_size[::-1]
    else:
        return page_size


@cli.command(group="Operations")
@click.argument("size", type=vp.PageSizeType(), required=True)
@click.option("-l", "--landscape", is_flag=True, default=False, help="Landscape orientation.")
@vp.global_processor
def pagesize(document: vp.Document, size, landscape) -> vp.Document:
    """Change the current page size.

    The page size is set (or modified) by the `read` command and used by the `write` command by
    default. This command can be used to set it to an arbitrary value. See the `write` command
    help section for more information on valid size value (`vpype write --help`).

    Note: this command only changes the current page size and has no effect on the geometries.
    Use the `translate` and `scale` commands to change the position and/or the scale of the
    geometries.

    Examples:

        Set the page size to A4:

            vpype [...] pagesize a4 [...]

        Set the page size to landscape A4:

            vpype [...] pagesize --landscape a4 [...]

        Set the page size to 11x15in:

            vpype [...] pagesize 11inx15in [...]
    """

    document.page_size = _normalize_page_size(size, landscape)
    return document


LAYOUT_HELP = f"""Layout the geometries on the provided page size.

SIZE may be one of:

    {', '.join(vp.PAGE_SIZES.keys())}

Alternatively, a custom size can be specified in the form of WIDTHxHEIGHT. WIDTH and
HEIGHT may include units. If only one has an unit, the other is assumed to have the
same unit. If none have units, both are assumed to be pixels by default. Here are some
examples:

\b
    --page-size 11x14in     # 11in by 14in
    --page-size 1024x768    # 1024px by 768px
    --page-size 13.5inx4cm  # 13.5in by 4cm

Portrait orientation is enforced, unless `--landscape` is used, in which case landscape
orientation is enforced.

By default, this command centers everything on the page. The horizontal and vertical
alignment can be adjusted using the `--align`, resp. `--valign` options.

Optionally, this command can scale the geometries to fit specified margins with the
`--fit-to-margins` option.

Examples:

\b
    Fit the geometries to 3cm margins with top alignment (a generally
    pleasing arrangement for square designs on portrait-oriented pages):

        vpype read input.svg layout --fit-to-margins 3cm --valign top a4 write.svg
"""


# noinspection PyShadowingNames
@cli.command(group="Operations", help=LAYOUT_HELP)
@click.argument("size", type=vp.PageSizeType(), required=True)
@click.option("-l", "--landscape", is_flag=True, default=False, help="Landscape orientation.")
@click.option(
    "-m",
    "--fit-to-margins",
    "margin",
    type=vp.LengthType(),
    help="Fit the geometries to page size with the specified margin.",
)
@click.option(
    "-h",
    "--align",
    type=click.Choice(["left", "center", "right"]),
    default="center",
    help="Horizontal alignment",
)
@click.option(
    "-v",
    "--valign",
    type=click.Choice(["top", "center", "bottom"]),
    default="center",
    help="Vertical alignment",
)
@vp.global_processor
def layout(
    document: vp.Document,
    size: Tuple[float, float],
    landscape: bool,
    margin: Optional[float],
    align: str,
    valign: str,
) -> vp.Document:
    """Layout command"""

    size = _normalize_page_size(size, landscape)

    document.page_size = size
    bounds = document.bounds()

    if bounds is None:
        # nothing to layout
        return document

    min_x, min_y, max_x, max_y = bounds
    width = max_x - min_x
    height = max_y - min_y
    if margin is not None:
        document.translate(-min_x, -min_y)
        scale = min((size[0] - 2 * margin) / width, (size[1] - 2 * margin) / height)
        document.scale(scale)
        min_x = min_y = 0.0
        width *= scale
        height *= scale
    else:
        margin = 0.0

    if align == "left":
        h_offset = margin - min_x
    elif align == "right":
        h_offset = size[0] - margin - width - min_x
    else:
        h_offset = margin + (size[0] - width - 2 * margin) / 2 - min_x

    if valign == "top":
        v_offset = margin - min_y
    elif valign == "bottom":
        v_offset = size[1] - margin - height - min_y
    else:
        v_offset = margin + (size[1] - height - 2 * margin) / 2 - min_y

    document.translate(h_offset, v_offset)
    return document


@cli.command(group="Operations")
@click.argument("pitch", type=vp.LengthType(), required=True)
@vp.layer_processor
def snap(line_collection: vp.LineCollection, pitch: float) -> vp.LineCollection:
    """Snap all points to a grid with with a spacing of PITCH.

    This command moves every point of every paths to the nearest grid point. If sequential
    points of a segment end up on the same grid point, they are deduplicated. If the resulting
    path contains less than 2 points, it is discarded.

    Example:

        Snap all points to a grid of 3x3mm:

            vpype [...] snap 3mm [...]
    """

    line_collection.scale(1 / pitch)

    new_lines = vp.LineCollection()
    for line in line_collection:
        new_line = np.round(line)
        idx = np.ones(len(new_line), dtype=bool)
        idx[1:] = new_line[1:] != new_line[:-1]
        if idx.sum() > 1:
            new_lines.append(np.round(new_line[idx]))

    new_lines.scale(pitch)
    return new_lines


@cli.command(group="Operations")
@vp.layer_processor
def reverse(line_collection: vp.LineCollection) -> vp.LineCollection:
    """Reverse order of lines.

    Reverse the order of lines within their respective layers. Individual lines are not
    modified (in particular, their trajectory is not inverted). Only the order in which they
    are drawn is reversed.
    """

    line_collection.reverse()
    return line_collection



"""This module implements a flexible, GPU-accelerated viewer for interactive and offscreen
rendering of :class:`vpype.Document` instances. It includes a Qt-based interactive backend as
well as a Pillow-based offscreen rendering backend.
"""
from ._scales import DEFAULT_SCALE_SPEC, UnitType
from .engine import *
from .image import ImageRenderer, render_image

# currently Qt is the only GUI backend so we unconditionally import from qt
from .qtviewer import QtViewer, show



"""Implementation of vpype's data model
"""
import logging
import math
from typing import Callable, Dict, Iterable, Iterator, List, Optional, Tuple, Union, cast

import numpy as np
from shapely.geometry import LinearRing, LineString, MultiLineString

from .geometry import crop, reloop
from .line_index import LineIndex

# REMINDER: anything added here must be added to docs/api.rst
__all__ = [
    "LineCollection",
    "Document",
    "LineLike",
    "LineCollectionLike",
    "as_vector",
    # deprecated:
    "VectorData",
]

LineLike = Union[LineString, LinearRing, Iterable[complex]]

# We accept LineString and LinearRing as line collection because MultiLineString are regularly
# converted to LineString/LinearRing when operation reduce them to single-line construct.
LineCollectionLike = Union[
    Iterable[LineLike], MultiLineString, "LineCollection", LineString, LinearRing
]


def as_vector(a: np.ndarray):
    """Return a view of a complex line array that behaves as an Nx2 real array"""
    return a.view(dtype=float).reshape(len(a), 2)


# noinspection PyShadowingNames
class LineCollection:
    """
    :py:class:`LineCollection` encapsulate a list of piecewise linear lines (or paths). Lines
    are implemented as 1D numpy arrays of complex numbers whose real and imaginary parts
    represent the X, respectively Y, coordinates of point in the paths.

    An instance of :py:class:`LineCollection` is used to model a single layer in vpype's
    :ref:`pipeline <fundamentals_pipeline>`. The complete pipeline is modelled by a
    :py:class:`Document` instance, which  essentially is a mapping of ``int`` (layer ID) to
    :py:class:`LineCollection`.

    Although the actual ``list`` is stored as private data member in :py:class:`LineCollection`
    instances, the class provides a sequence API similar to ``list``::

        >>> import vpype, numpy as np
        >>> lc = vpype.LineCollection()
        >>> lc.append(np.array([0, 10. + 10.j]))
        >>> lc.append(np.array([10.j, 5. + 5.j]))
        >>> len(lc)
        2
        >>> lc[0]
        array([ 0. +0.j, 10.+10.j])
        >>> for line in lc:
        ...    print(repr(line))
        ...
        array([ 0. +0.j, 10.+10.j])
        array([0.+10.j, 5. +5.j])

    In addition to Numpy arrays, the class accepts paths expressed in a variety of format
    including Python ``list`` or Shapely objects::

        >>> from shapely.geometry import LineString, LinearRing, MultiLineString
        >>> lc = vpype.LineCollection()
        >>> lc.append([5, 5+5j])
        >>> lc.append(LineString([(1, 1), (3, 2)]))
        >>> lc.append(LinearRing([(0, 0), (1, 0), (1, 1), (0, 1)]))
        >>> lc.extend(MultiLineString([[(0, 0), (10, 0)], [(4, 4), (0, 4)]]))
        >>> lc
        LineCollection([array([5.+0.j, 5.+5.j]), array([1.+1.j, 3.+2.j]), array([0.+0.j,
        1.+0.j, 1.+1.j, 0.+1.j, 0.+0.j]), array([ 0.+0.j, 10.+0.j]), array([4.+4.j, 0.+4.j])])

    Instances can also be converted to Shapely's MultiLineString:

        >>> mls = lc.as_mls()
        >>> print(mls)
        MULTILINESTRING ((5 0, 5 5), (1 1, 3 2), (0 0, 1 0, 1 1, 0 1, 0 0), (0 0, 10 0),
        (4 4, 0 4))

    Finally, :py:class:`LineCollection` implements a number of operations such as geometrical
    transformation, cropping, merging, etc. (see member function documentation for details).
    """

    def __init__(self, lines: LineCollectionLike = ()):
        """Create a LineCollection instance from an iterable of lines.

        Args:
            lines (LineCollectionLike): iterable of line (accepts the same input as
                :func:`~LineCollection.append`).
        """
        self._lines: List[np.ndarray] = []

        self.extend(lines)

    @property
    def lines(self) -> List[np.ndarray]:
        """Returns the list of line.

        Returns:
            list of line
        """
        return self._lines

    def append(self, line: LineLike) -> None:
        """Append a single line.

        This function accepts an iterable of complex or a Shapely geometry
        (:py:class:`LineString` or :py:class:`LinearRing`).

        Args:
            line (LineLike): line to append
        """
        if isinstance(line, LineString) or isinstance(line, LinearRing):
            # noinspection PyTypeChecker
            self._lines.append(np.array(line).view(dtype=complex).reshape(-1))
        else:
            line = np.array(line, dtype=complex).reshape(-1)
            if len(line) > 1:
                self._lines.append(line)

    def extend(self, lines: LineCollectionLike) -> None:
        """Append lines from a collection.

        This function accepts an iterable of iterable of complex, another
        :py:class:`LineCollection` instance, or a Shapely geometry
        (:py:class:`MultiLineString`, :py:class:`LineString` or :py:class:`LinearRing`).

        Shapely's LineString and LinearRing are occasionally obtained when a MultiLineString is
        actually expected. As a result, they are accepted as input even though they are not,
        strictly speaking, a line collection.

        Args:
            lines (LineCollectionLike): lines to append
        """

        if hasattr(lines, "geom_type") and lines.is_empty:  # type: ignore
            return

        # sometimes, mls end up actually being ls
        if isinstance(lines, LineString) or isinstance(lines, LinearRing):
            lines = [lines]

        for line in lines:
            self.append(line)

    def is_empty(self) -> bool:
        """Check for emptiness.

        Returns:
            True if the instance does not contain any line, False otherwise.
        """
        return len(self) == 0

    def reverse(self) -> None:
        """Reverse order of the lines."""
        self._lines = list(reversed(self._lines))

    def __iter__(self):
        return self._lines.__iter__()

    def __len__(self) -> int:
        return len(self._lines)

    def __getitem__(self, item: Union[int, slice]):
        return self._lines[item]

    def __repr__(self):
        return f"LineCollection({self._lines})"

    def as_mls(self) -> MultiLineString:
        """Converts the LineCollection to a :py:class:`MultiLineString`.

        Returns:
            a MultiLineString Shapely object
        """
        return MultiLineString([as_vector(line) for line in self.lines])

    def translate(self, dx: float, dy: float) -> None:
        """Translates all line by a given offset.

        Args:
            dx: offset along X axis
            dy: offset along Y axis
        """
        c = complex(dx, dy)
        for line in self._lines:
            line += c

    def scale(self, sx: float, sy: Optional[float] = None) -> None:
        """Scale the geometry.

        The scaling is performed about the coordinates origin (0, 0). To scale around a
        specific location, appropriate translations must be performed before and after the
        scaling::

            >>> import vpype
            >>> lc = vpype.LineCollection([(-1+1j, 1+1j)])
            >>> lc.translate(0, -1)
            >>> lc.scale(1.2)
            >>> lc.translate(0, 1)
            >>> lc
            LineCollection([array([-1.2+1.j,  1.2+1.j])])

        Args:
            sx: scale factor along x
            sy: scale factor along y (if None, then sx is used)
        """
        if sy is None:
            sy = sx

        for line in self._lines:
            line.real *= sx
            line.imag *= sy

    def rotate(self, angle: float) -> None:
        """Rotates the geometry by ``angle`` amount.

        The angle is expressed in radian. Positive value rotate clockwise.

        The rotation is performed about the coordinates origin (0, 0). To rotate around a
        specific location, appropriate translations must be performed before and after the
        scaling::

            >>> import vpype
            >>> lc = vpype.LineCollection([(-1+1j, 1+1j)])
            >>> lc.translate(0, -1)
            >>> lc.rotate(1.2)
            >>> lc.translate(0, 1)

        Args:
            angle: rotation angle in rad
        """
        c = complex(math.cos(angle), math.sin(angle))
        for line in self._lines:
            line *= c

    def skew(self, ax: float, ay: float) -> None:
        """Skew the geometry by some angular amounts along X and Y axes.

        The angle is expressed in radians.

        The skew is performed about the coordinates origin (0, 0). To rotate around a
        specific location, appropriate translations must be performed before and after the
        scaling::

            >>> import vpype
            >>> lc = vpype.LineCollection([(-1+1j, 1+1j)])
            >>> lc.translate(0, -1)
            >>> lc.skew(0., 1.2)
            >>> lc.translate(0, 1)

        Args:
            ax: skew angle in rad along X axis
            ay: skew angle in rad along Y axis
        """
        tx, ty = math.tan(ax), math.tan(ay)
        for line in self._lines:
            line += tx * line.imag + 1j * ty * line.real

    def reloop(self, tolerance: float) -> None:
        """Randomizes the seam of closed paths. Paths are considered closed when their first
        and last point are closer than *tolerance*.

        :param tolerance: tolerance to determine if a path is closed
        """

        for i, line in enumerate(self._lines):
            delta = line[-1] - line[0]
            if np.hypot(delta.real, delta.imag) <= tolerance:
                self._lines[i] = reloop(line)

    def crop(self, x1: float, y1: float, x2: float, y2: float) -> None:
        """Crop all lines to a rectangular area.

        Args:
            x1, y1: first corner of the crop area
            x2, y2: second corner of the crop area
        """

        if x1 > x2:
            x1, x2 = x2, x1
        if y1 > y2:
            y1, y2 = y2, y1

        if x1 == x2 or y1 == y2:
            self._lines = []
        else:
            new_lines = []
            for line in self._lines:
                new_lines.extend(crop(line, x1, y1, x2, y2))
            self._lines = new_lines

    def filter(self, key: Callable[[np.ndarray], bool]) -> None:
        """Remove lines from the :class:`LineCollection` for which key returns False.

        Args:
            key: filter (returns True if the line should be kept or False otherwise)
        """
        self._lines = [line for line in self._lines if key(line)]

    def merge(self, tolerance: float, flip: bool = True) -> None:
        """Merge lines whose endings overlap or are very close.

        Args:
            tolerance: max distance between line ending that may be merged
            flip: allow flipping line direction for further merging
        """
        if len(self) < 2:
            return

        index = LineIndex(self.lines, reverse=flip)
        new_lines = LineCollection()

        while len(index) > 0:
            line = index.pop_front()

            # we append to `line` until we dont find anything to add
            while True:
                idx, reverse = index.find_nearest_within(line[-1], tolerance)
                if idx is None and flip:
                    idx, reverse = index.find_nearest_within(line[0], tolerance)
                    line = np.flip(line)
                if idx is None:
                    break
                new_line = cast(np.ndarray, index.pop(idx))
                if reverse:
                    new_line = np.flip(new_line)
                line = np.hstack([line, new_line])

            new_lines.append(line)

        self._lines = new_lines._lines

    def bounds(self) -> Optional[Tuple[float, float, float, float]]:
        """Returns the geometries' bounding box.

        Returns:
            tuple (xmin, ymin, xmax, ymax) for the bounding box or None if the LineCollection
            is empty
        """
        if len(self._lines) == 0:
            return None
        else:
            return (
                float(min((line.real.min() for line in self._lines))),
                float(min((line.imag.min() for line in self._lines))),
                float(max((line.real.max() for line in self._lines))),
                float(max((line.imag.max() for line in self._lines))),
            )

    def width(self) -> float:
        """Returns the total width of the geometries.

        Returns:
            the width (xmax - xmin) or 0.0 if the LineCollection is empty
        """

        if self._lines:
            return float(
                max((line.real.max() for line in self._lines))
                - min((line.real.min() for line in self._lines))
            )
        else:
            return 0.0

    def height(self) -> float:
        """Returns the total height of the geometries.

        Returns:
            the width (ymax - ymin) or 0.0 if the LineCollection is empty
        """
        if self._lines:
            return float(
                max((line.imag.max() for line in self._lines))
                - min((line.imag.min() for line in self._lines))
            )
        else:
            return 0.0

    def length(self) -> float:
        """Return the total length of the paths.

        Returns:
            the total length
        """
        return sum(np.sum(np.abs(np.diff(line))) for line in self._lines)

    def pen_up_trajectories(self) -> "LineCollection":
        """Returns a LineCollection containing the pen-up trajectories."""
        return LineCollection(
            ([self._lines[i][-1], self._lines[i + 1][0]] for i in range(len(self._lines) - 1)),
        )

    def pen_up_length(self) -> Tuple[float, float, float]:
        """Returns statistics on the pen-up distance corresponding to the path.

        The total, mean, and median distance are returned. The pen-up distance is the distance
        between a path's end and the next path's beginning.

        Returns:
            tuple (total, mean, median) for the pen-up distances
        """
        if len(self.lines) < 2:
            return 0.0, 0.0, 0.0

        ends = np.array([line[-1] for line in self.lines[:-1]])
        starts = np.array([line[0] for line in self.lines[1:]])
        dists = np.abs(starts - ends)
        # noinspection PyTypeChecker
        return float(np.sum(dists)), float(np.mean(dists)), float(np.median(dists))

    def segment_count(self) -> int:
        """Returns the total number of segment across all lines.

        Returns:
            the total number of segments in the geometries
        """
        return sum(max(0, len(line) - 1) for line in self._lines)


class Document:
    """This class is the core data model of vpype and represent the data that is passed from
    one command to the other. At its core, a Document is a collection of layers identified
    by non-zero positive integers and each represented by a :py:class:`LineCollection`.

    In addition, the Document class maintains a :py:attr:`page_size` attribute which describe
    the physical size of the document. This attribute is not strictly linked to the actual
    Document's content, but can be set based on it.
    """

    def __init__(
        self,
        line_collection: LineCollection = None,
        page_size: Optional[Tuple[float, float]] = None,
    ):
        """Create a Document, optionally providing a :py:class:`LayerCollection` for layer 1.

        Args:
            line_collection: if provided, used as layer 1
        """
        self._layers: Dict[int, LineCollection] = {}
        self._page_size: Optional[Tuple[float, float]] = page_size

        if line_collection:
            self.add(line_collection, 1)

    def empty_copy(self) -> "Document":
        """Create an empty copy of this document with the same page size."""
        return Document(page_size=self.page_size)

    @property
    def layers(self) -> Dict[int, LineCollection]:
        """Returns a reference to the layer dictionary.
        Returns:
            the internal layer dictionary
        """
        return self._layers

    @property
    def page_size(self) -> Optional[Tuple[float, float]]:
        """Returns the page size or None if it hasn't been set."""
        return self._page_size

    @page_size.setter
    def page_size(self, page_size=Optional[Tuple[float, float]]) -> None:
        """Sets the page size to a new value."""
        self._page_size = page_size

    def extend_page_size(self, page_size: Optional[Tuple[float, float]]) -> None:
        """Adjust the  page sized according to the following logic:

            - if ``page_size`` is None, the the page size is unchanged
            - if ``self.page_size`` is None, it is set to ``page_size``
            - if both page sizes are not None, the page size is set to the largest value in
              both direction

        Args:
            page_size: page dimension to use to update ``self.page_size``
        """
        if page_size:
            if self.page_size:
                self.page_size = (
                    max(self.page_size[0], page_size[0]),
                    max(self.page_size[1], page_size[1]),
                )
            else:
                self.page_size = page_size

    def ids(self) -> Iterable[int]:
        """Returns the list of layer IDs"""
        return self._layers.keys()

    def layers_from_ids(self, layer_ids: Iterable[int]) -> Iterator[LineCollection]:
        """Returns an iterator that yield layers corresponding to the provided IDs, provided
        they exist. This is typically used to process a command's layer list option, in
        combination with :py:func:`multiple_to_layer_ids`.

        Non-existent layer IDs in the input are ignored.

        Args:
            layer_ids: iterable of layer IDs

        Returns:
            layer iterator
        """
        return (self._layers[lid] for lid in layer_ids if lid in self._layers)

    def exists(self, layer_id: int) -> bool:
        """Test existence of a layer.

        Note that existence of a layer does not necessarily imply that it isn't empty.

        Args:
            layer_id: layer ID to test

        Returns:
            True if the layer ID exists
        """
        return layer_id in self._layers

    def __getitem__(self, layer_id: int):
        return self._layers.__getitem__(layer_id)

    def __setitem__(self, layer_id: int, value: LineCollectionLike):
        if layer_id < 1:
            raise ValueError(f"expected non-null, positive layer id, got {layer_id} instead")

        if isinstance(value, LineCollection):
            self._layers[layer_id] = value
        else:
            self._layers[layer_id] = LineCollection(value)

    def free_id(self) -> int:
        """Returns the lowest unused layer id.

        Returns:
            the unused layer ID
        """
        vid = 1
        while vid in self._layers:
            vid += 1
        return vid

    def add(self, lc: LineCollection, layer_id: Union[None, int] = None) -> None:
        """Add a the content of a :py:class:`LineCollection` to a given layer.

        If the given layer is None, the input LineCollection is used to create a new layer
        using the lowest available layer ID.
        """
        if layer_id is None:
            layer_id = 1
            while layer_id in self._layers:
                layer_id += 1

        if layer_id in self._layers:
            self._layers[layer_id].extend(lc)
        else:
            self._layers[layer_id] = lc

    def extend(self, doc: "Document") -> None:
        """Extend a Document with the content of another Document.

        The layer structure of the source Document is maintained and geometries are either
        appended to the destination's corresponding layer or new layers are created, depending
        on if the layer existed or not in the destination Document.

        The :py:attr:`page_size` attribute is adjusted using :meth:`extend_page_size`.

        Args:
            doc: source Document
        """

        self.extend_page_size(doc.page_size)

        for layer_id, layer in doc.layers.items():
            self.add(layer, layer_id)

    def is_empty(self) -> bool:
        """Returns True if all layers are empty.

        Returns:
            True if all layers are empty"""
        for layer in self.layers.values():
            if not layer.is_empty():
                return False
        return True

    def pop(self, layer_id: int) -> LineCollection:
        """Removes a layer from the Document.

        Args:
            layer_id: ID of the layer to be removed

        Returns:
            the :py:class:`LineCollection` corresponding to the removed layer
        """
        return self._layers.pop(layer_id)

    def count(self) -> int:
        """Returns the total number of layers.

        Returns:
            total number of layer"""
        return len(self._layers.keys())

    def translate(self, dx: float, dy: float) -> None:
        """Translates all line by a given offset.

        Args:
            dx: offset along X axis
            dy: offset along Y axis
        """
        for layer in self._layers.values():
            layer.translate(dx, dy)

    def scale(self, sx: float, sy: Optional[float] = None) -> None:
        """Scale the geometry.

        The scaling is performed about the coordinates origin (0, 0). To scale around a
        specific location, appropriate translations must be performed before and after the
        scaling (see :func:`LineCollection.scale`).

        Args:
            sx: scale factor along x
            sy: scale factor along y (if None, then sx is used)
        """
        for layer in self._layers.values():
            layer.scale(sx, sy)

    def rotate(self, angle: float) -> None:
        """Rotate the Document's content..

        The rotation is performed about the coordinates origin (0, 0). To rotate around a
        specific location, appropriate translations must be performed before and after the
        scaling (see :func:`LineCollection.rotate`).

        Args:
            angle: rotation angle (radian)
        """
        for layer in self._layers.values():
            layer.rotate(angle)

    def bounds(
        self, layer_ids: Union[None, Iterable[int]] = None
    ) -> Optional[Tuple[float, float, float, float]]:
        """Compute bounds of the document.

        If layer_ids is provided, bounds are computed only for the corresponding IDs.

        Note: the bounds are computed based on the actual geometries contained in this
        :class:`Document` instance. The document's page size, if any, is not taken into account
        by this calculation.

        Args:
            layer_ids: layers to consider in the bound calculation

        Returns:
            boundaries of the geometries
        """
        if layer_ids is None:
            layer_ids = self.ids()
        a = np.array(
            [
                self._layers[vid].bounds()
                for vid in layer_ids
                if self.exists(vid) and len(self._layers[vid]) > 0
            ]
        )
        if len(a) > 0:
            return a[:, 0].min(), a[:, 1].min(), a[:, 2].max(), a[:, 3].max()
        else:
            return None

    def crop(self, x1: float, y1: float, x2: float, y2: float) -> None:
        """Crop all layers to a rectangular area.

        Args:
            x1, y1: first corner of the crop area
            x2, y2: second corner of the crop area
        """
        for layer in self._layers.values():
            layer.crop(x1, y1, x2, y2)

    def fit_page_size_to_content(self) -> None:
        """Set :py:attr:`page_size` to the current geometries' width and height and move
        the geometries so that their bounds align to (0, 0).
        """
        bounds = self.bounds()
        if not bounds:
            return

        self.translate(-bounds[0], -bounds[1])
        self.page_size = (bounds[2] - bounds[0], bounds[3] - bounds[1])

    def length(self) -> float:
        """Return the total length of the paths.

        Returns:
            the total length
        """
        return sum(layer.length() for layer in self._layers.values())

    def pen_up_length(self) -> float:
        """Returns the total pen-up distance corresponding to the path.

        This function does not account for the pen-up distance between layers.

        Returns:
            total pen-up distances
        """
        return sum(layer.pen_up_length()[0] for layer in self._layers.values())

    def segment_count(self) -> int:
        """Returns the total number of segment across all lines.

        Returns:
            the total number of segments in the geometries
        """
        return sum(layer.segment_count() for layer in self._layers.values())


class VectorData(Document):
    """Deprecated, use Document."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        logging.warning("!!! `vpype.VectorData` is deprecated, use `vpype.Document` instead.")



import math

import numpy as np

__ALL__ = ["line", "rect", "arc", "circle", "ellipse"]


def line(x0: float, y0: float, x1: float, y1: float) -> np.ndarray:
    """Build a line from two points

    Args:
        x0: line start X coordinate
        y0: line start Y coordinate
        x1: line end X coordinate
        y1: line end Y coordinate

    Returns:
        line path
    """
    return np.array([complex(x0, y0), complex(x1, y1)])


def rect(
    x: float,
    y: float,
    width: float,
    height: float,
    tl: float = 0,
    tr: float = 0,
    br: float = 0,
    bl: float = 0,
    quantization: float = 0.1,
) -> np.ndarray:
    """Build a rectangular path, with optional rounded angles.

    Args:
        x: top-left corner X coordinate
        y: top-left corner Y coordinate
        width: rectangle width
        height: rectangle height
        tl: top-left corner radius (0 if not provided)
        tr: top-right corner radius (0 if not provided)
        br: bottom-right corner radius (0 if not provided)
        bl: bottom-left corner radius (0 if not provided)
        quantization: maximum size of segments approximating round corners

    Returns:
        rectangular path
    """
    if (tr + tl) > width:
        scale = width / (tr + tl)
        tr *= scale
        tl *= scale
    if (br + bl) > width:
        scale = width / (br + bl)
        tr *= scale
        tl *= scale
    if (tl + bl) > height:
        scale = height / (tl + bl)
        tl *= scale
        bl *= scale
    if (tr + br) > height:
        scale = height / (tr + br)
        tr *= scale
        br *= scale

    if tl != 0:
        p1 = arc(x + tl, y + tl, tl, tl, 90, 180, quantization)
    else:
        p1 = np.array([complex(x, y)])

    if bl != 0:
        p2 = arc(x + bl, y + height - bl, bl, bl, 180, 270, quantization)
    else:
        p2 = np.array([complex(x, y + height)])

    if br != 0:
        p3 = arc(x + width - br, y + height - br, br, br, 270, 360, quantization)
    else:
        p3 = np.array([complex(x + width, y + height)])

    if tr != 0:
        p4 = arc(x + width - tr, y + tr, tr, tr, 0, 90, quantization)
    else:
        p4 = np.array([complex(x + width, y)])

    return np.concatenate((p1, p2, p3, p4, np.array([p1[0]])))


def arc(
    x: float,
    y: float,
    rw: float,
    rh: float,
    start: float,
    stop: float,
    quantization: float = 0.1,
) -> np.ndarray:
    """Build an elliptical arc path. Zero angles refer to east of unit circle and positive
    values extend counter-clockwise.

    Args:
        x: center X coordinate
        y: center Y coordinate
        rw: ellipse half-width
        rh: ellipse half-height (use the same value as ``rw`` for a circular arc)
        start: start angle (degree)
        stop: stop angle (degree)
        quantization: maximum length of linear segment

    Returns:
        arc path
    """

    def normalize_angle(a):
        while a > 360:
            a -= 360
        while a < 0:
            a += 360
        return a

    start = normalize_angle(start)
    stop = normalize_angle(stop)
    if stop < start:
        stop += 360
    elif start == stop:
        raise ValueError("start and stop angles must have different values")

    n = max(3, math.ceil((stop - start) / 180 * math.pi * max(rw, rh) / quantization))
    angle = np.linspace(start, stop, n + 1)
    angle[angle == 360] = 0
    angle *= math.pi / 180
    return rw * np.cos(-angle) + 1j * rh * np.sin(-angle) + complex(x, y)


def circle(x: float, y: float, radius: float, quantization: float = 0.1) -> np.ndarray:
    """Build a circular path.

    Args:
        x: center X coordinate
        y: center Y coordinate
        radius: circle radius
        quantization: maximum length of linear segment

    Returns:
        circular path
    """
    return arc(x, y, radius, radius, 0, 360, quantization)


def ellipse(x: float, y: float, w: float, h: float, quantization: float = 0.1) -> np.ndarray:
    """Build an elliptical path.

    Args:
        x: center X coordinate
        y: center Y coordinate
        w: half width of the ellipse
        h: half height of the ellipse
        quantization: maximum length of linear segment

    Returns:
        elliptical path
    """
    return arc(x, y, w, h, 0, 360, quantization)
